# -*- coding: utf-8 -*-
"""
Created on Wed Jul  4 10:03:52 2012

@author: dabi
"""

from numpy import *
from scipy import stats
from matplotlib.pyplot import *

import Spectral1D as spec1D
import UncertaintyQuantification

close('all')

#dists = [stats.uniform(), stats.norm(0,1), stats.norm(2,3), stats.norm(2,3), stats.norm(2,3)]
#polys = [spec1D.Poly1D(spec1D.JACOBI,(0,0)), spec1D.Poly1D(spec1D.HERMITEP_PROB,None), spec1D.Poly1D(spec1D.HERMITEP_PROB,None), spec1D.Poly1D(spec1D.HERMITEP_PROB,None), spec1D.Poly1D(spec1D.HERMITEP_PROB,None)]
#distsPoly = [stats.uniform(), stats.norm(0,1), stats.norm(0,1), stats.norm(0,1), stats.norm(0,1)]
#Ns = [3,3,4,5,10]

#dists = [stats.uniform(), stats.norm(2,1)]
#polys = [spec1D.Poly1D(spec1D.JACOBI,(0,0)), spec1D.Poly1D(spec1D.HERMITEP_PROB,None)]
#distsPoly = [stats.uniform(-1,2), stats.norm(0,1)]
#Ns = [8,10]

dists = [stats.norm(-3,4), stats.norm(2,1)]
polys = [spec1D.Poly1D(spec1D.HERMITEP_PROB,None), spec1D.Poly1D(spec1D.HERMITEP_PROB,None)]
distsPoly = [stats.norm(0,1), stats.norm(0,1)]
Ns = [3,2]

dictionary = UncertaintyQuantification.PCM_Tensor(dists,polys,distsPoly,Ns,True)

vals = asarray(dictionary['vals'])
w = dictionary['wCub']
wPrj = dictionary['wPrj']
V = dictionary['V']

# Plot collocation points
figure()
plot(vals[:,0],vals[:,1],'.')


#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

__all__ = ['RunUnitTests','RunTestKL1D','RunTestKL2D','RunTestRandomSampling','RunTestTimingPreFit','RunTestCutANOVA']

def RunTestRandomSampling(maxprocs=None,PLOTTING=False):
    from UQToolbox.unittests import TestRandomSampling
    TestRandomSampling.run(maxprocs,PLOTTING)

def RunTestKL1D(maxprocs=None,PLOTTING=False):
    from UQToolbox.unittests import TestKL1D
    TestKL1D.run(maxprocs,PLOTTING)

def RunTestKL2D(maxprocs=None,PLOTTING=False):
    from UQToolbox.unittests import TestKL2D
    TestKL2D.run(maxprocs,PLOTTING)

def RunTestTimingPreFit(maxprocs=None,PLOTTING=False):
    from UQToolbox.unittests import TestTimingPreFit
    TestTimingPreFit.run(maxprocs,PLOTTING)

def RunTestCutANOVA(maxprocs=None,PLOTTING=False):
    from UQToolbox.unittests import TestCutANOVA
    TestCutANOVA.run(maxprocs,PLOTTING)

def RunUnitTests(maxprocs=None,PLOTTING=False):
    """ Runs all the unit tests.
    
    :params int maxprocs: If MPI support is enabled, defines how many processors to use.
    """
    RunTestRandomSampling(maxprocs,PLOTTING)
    RunTestCutANOVA(maxprocs,PLOTTING)
    RunTestKL1D(maxprocs,PLOTTING)
    RunTestKL2D(maxprocs,PLOTTING)



#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[1;32m'
    WARNING = '\033[93m'
    FAIL = '\033[1;91m'
    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

import sys
import numpy as np

def MetropolisHastingsWithinGibbs(n, nMC, X0, target, propRG, propPDF):
    # n is the length of the Markov Chains
    # nMC is the number of Markov Chains (n. of samples)
    # target is the target function returning the log-likelihood
    # propRG is the random number generator for the proposed distribution
    # propPDF is the PDF of the proposed distribution
    from numpy.random import binomial as rbinom
    X = np.empty((n,nMC),dtype=np.float)
    X[0,:] = X0
    for t in range(0,n-1):
        XBase = X[t,:]
        for i in range(0,nMC):
            XNew = XBase.copy()
            XNew[i] = propRG(XBase[i],1)
            logR = ( target(XNew) + propPDF(XNew[i],XBase[i]) ) - ( target(XBase) + propPDF(XBase[i],XNew[i]))
            if logR > 0 :
                X[t+1,i] = XNew[i]
            else:
                # Random acceptance of step
                if rbinom(n=1,p=np.min((1,np.exp(logR))),size=1)[0]:
                    X[t+1,i] = XNew[i]
                else:
                    X[t+1,i] = X[t,i]
            if i < nMC:
                XBase = np.hstack( (X[t+1,:i+1] , X[t,i+1:]) )
    return X

def MetropolisHastings(n, nMC, X0, target, propRG, propPDF):
    # n is the length of the Markov Chains
    # nMC is the number of Markov Chains (n. of samples)
    # target is the target function returning the log-likelihood
    # propRG is the random number generator for the proposed distribution
    # propPDF is the PDF of the proposed distribution
    from numpy.random import binomial as rbinom
    X = np.empty((n,nMC),dtype=np.float)
    X[0,:] = X0
    for t in range(0,n-1):
        sys.stderr.write("MH Iteration: %d/%d\r" % (t,n))
        sys.stderr.flush()
        XBase = X[t,:]
        XNew = np.array([propRG(XBase[0],1)[0],propRG(XBase[1],1)[0]])
        logR = ( target(XNew) + propPDF(XNew[0],XBase[0]) + propPDF(XNew[1],XBase[1]) ) - ( target(XBase) + propPDF(XBase[0],XNew[0]) + propPDF(XBase[1],XNew[1]))
        if logR > 0 :
            X[t+1,:] = XNew
        else:
            # Random acceptance of step
            if rbinom(n=1,p=np.min((1,np.exp(logR))),size=1)[0]:
                X[t+1,:] = XNew
            else:
                X[t+1,:] = X[t,:]
    sys.stderr.write("\t")
    sys.stderr.flush()
    return X



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#


import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats
import UQToolbox.RandomSampling as RS

params = dict( tspan = [0.,1.],
               nt = 100 )
    
def f(X,params):
    # y(t) = b * exp(-a*t)
    # In the following X[0] = a, X[1] = b
    # Note: all the "imports" must be declared inside this function
    # in order for UQToolbox to work with mpi
    import numpy as np
    tspan = params['tspan']
    nt = params['nt']
    T = np.linspace( tspan[0], tspan[1], nt )
    return X[1] * np.exp( - X[0] * T )


# Optional storage information
store_file = 'rs.pkl'
store_freq = 100 # Every how many simulations to store

# Optional running information
maxprocs = 1 # number of mpi threads to be used

# Distribution of (a,b)
dists = [ stats.uniform(.5,.5),    # a ~ U([0.5, 1.0])
          stats.uniform(.25,.5) ]  # b ~ U([0.25, 0.75])

# Number of experiments
N = 1000

# The experiment object
experiments = RS.RandomExperiments( f, params,
                                    dists,
                                    marshal_f = True, # Store definition of f
                                    store_file = store_file )

# Test Sobol sequence
experiments.sample( N, 'sobol' ) # Create the qmc samples. Other options are 'mc' or 'lhc'

# Plot samples
ns = np.asarray( experiments.get_new_samples() ) # Sample points on which the function has not been evaluated yet
plt.figure()
plt.plot( ns[:,0], ns[:,1], '.' )
plt.show(False)

print "N samples to be evaluated: %d" % len( experiments.get_new_samples() )

# Evaluate function on samples
print "Running experiments"
experiments.run( maxprocs, store_freq=store_freq )

print "N samples to be evaluated: %d" % len( experiments.get_new_samples() )
print "N samples already evaluated: %d" % len( experiments.get_samples() )

# Evaluate mean and variance
results = np.asarray( experiments.get_results() )
mean = np.mean( results, axis=0 )
var = np.var( results, axis=0 )

# Plot mean, var and first 2 realizations
tspan = params['tspan']
nt = params['nt']
T = np.linspace( tspan[0], tspan[1], nt )

plt.figure()
for i in range(2):
    plt.plot(T, results[i,:])

plt.plot(T, mean, 'k', linewidth=2.,label='mean')
plt.plot(T, mean + np.sqrt(var), '--k', linewidth=2.,label='std')
plt.plot(T, mean - np.sqrt(var), '--k', linewidth=2.)
plt.legend()
plt.show(False)



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

import numpy as np

from numpy import linalg as la

from scipy import stats

import itertools

from prettytable import PrettyTable

from SpectralToolbox import Spectral1D
from SpectralToolbox import SpectralND
from SpectralToolbox import SparseGrids

def gPC_MultiIndex(dist,poly,distPoly,N,NI=[],quadTypes=None,left=None,right=None,end=None,warnings=True):
    """
    gPC_MultiIndex(dist,poly,distPoly,N)
    Generates the Multi Index basis Vandermonde matrix using the list of 
    distributions of the parameters space approximated by the selected polynomials.

    :param dist: list of n distributions representing the n-Dimensional random space
    :param poly: list of n polynomials (class Spectral1D) used for approximating each of the directions in the random space
    :param distPoly: list of n standard distributions associated to the polynomial selected
    :param N: maximum polynomial order in the basis
    :param NI: (int) number of points in which to evaluate the basis functions for precise cubature rules (if NI=[], the interpolated vandermonde is not evaluated)
    :param quadTypes: (list, ``Spectral1D.AVAIL_POINTS``) list of n types of quadrature points among Gauss, Gauss-Lobatto and Gauss-Radau. If ``None``, Gauss quadrature points are used.
    :param left: (list,float) list of left values used by ORTHPOL for Gauss-Lobatto rules (the dimensions where the value is not used can be set to anything)
    :param right: (list,float) list of left values used by ORTHPOL for Gauss-Lobatto rules (the dimensions where the value is not used can be set to anything)
    :param end: (list,float) list of left values used by ORTHPOL for Gauss-Radau rules (the dimensions where the value is not used can be set to anything)
    :param warnings: True if you want to read warning on memory allocation
    
    :returns: dictionary with the following attributes
        ``x``: tensor product of collocation points in the standard distributions
        ``w``: tensor product of the weights for cubature/projection rules (sum to 1 or prod(gamma0) respectively)
        ``vals``: tensor product of collocation points
        ``V``: Pascal's simplex of the basis functions
    """
    
    if (len(dist) != len(distPoly) or len(distPoly) != len(poly)) :
        print("The number of elements in dist, distPoly and poly is not consistent")
        return
    else:
        DIM = len(dist)

    if quadTypes == None:
        quadTypes = [Spectral1D.GAUSS for i in range(DIM)]
    if left == None:
        left = [None for i in range(DIM)]
    if right == None:
        right = [None for i in range(DIM)]
    if end == None:
        end = [None for i in range(DIM)]

    if DIM != len(quadTypes) or DIM != len(left) or DIM != len(right) or DIM != len(end):
        print("The number of elements in quadTypes, left, right, end are not consistent")
        return

    pp = SpectralND.PolyND(poly)
    
    (x, w) = pp.Quadrature(np.tile(N,DIM),quadTypes=quadTypes,norm=False, left=left, right=right, end=end)
    
    # Compute relative samples using inverse CDF method
    vals = np.zeros(np.asarray(x).shape)
    for i in range(0,DIM):
        vals[:,i] = dist[i].ppf(distPoly[i].cdf(x[:,i]))
    
    # Compute Generalized Vandermonde if required
    xs = []
    for i in range(0,DIM):
        x1D, _ = poly[i].Quadrature(N,quadType=quadTypes[i],norm=False,left=left[i], right=right[i], end=end[i])
        xs.append(x1D)
    V = pp.GradVandermondePascalSimplex(xs, N, np.zeros(DIM,dtype=int), norms=None, warnings=warnings)
    
    # Compute interpolated values of the Vandermonde matrix
    (xI, wI) = pp.Quadrature(np.tile(NI,DIM),quadTypes=quadTypes,norm=False, left=left, right=right, end=end)
    
    xs = []
    for i in range(0,DIM):
        x1D, _ = poly[i].Quadrature(NI,quadType=quadTypes[i],norm=False,left=left[i], right=right[i], end=end[i])
        xs.append(x1D)
    VI = pp.GradVandermondePascalSimplex(xs, N, np.zeros(DIM,dtype=int), norms=None)

    IDX = SpectralND.MultiIndex(DIM,N)
    
    # Create Dictionary
    dictionary = {'x':    x,
                  'w':    w,
                  'vals': vals,
                  'V':    V,
                  'xI':   xI,
                  'wI':   wI,
                  'VI':   VI,
                  'IDX':  IDX}
    
    return dictionary

def PCM_Tensor(dist,poly,distPoly,Ns,allocBasis=True):
    """
    PCM_Tensor(dist,poly,distPoly,N)
    Generates the full tensor grid using the list of distributions of the parameters
    space approximated by the selected polynomials.

    :param dist: list of n distributions representing the n-Dimensional random space
    :param poly: list of n polynomials (class Spectral1D) used for approximating each of the directions in the random space
    :param distPoly: list of n standard distributions associated to the polynomial selected
    :param Ns: list of n integers for the order of polynomial expansion in each direction
    :param allocBasis: boolean. True if you want the tensor basis.
    
    :returns: dictionary with the following attributes
        'x': tensor product of collocation points in the standard distributions
        'w': tensor product of the weights for cubature/projection rules (sum to 1 or prod(gamma0) respectively)
        'vals': tensor product of collocation points
        'V': tensor basis functions
    """
    
    if (len(dist) != len(distPoly) or len(distPoly) != len(poly) or len(poly) != len(Ns)) :
        print("The number of elements in dist, distPoly and poly is not consistent")
        return
    else:
        DIM = len(dist)

    pp = SpectralND.PolyND(poly)

    (x, w) = pp.GaussQuadrature(Ns,norm=False)
    nrmw = w/np.sum(w)
    
    # Compute relative samples using inverse CDF method
    vals = np.zeros(np.asarray(x).shape)
    for i in range(0,DIM):
        vals[:,i] = dist[i].ppf(distPoly[i].cdf(x[:,i]))
    
    # Compute Generalized Vandermonde if required
    if allocBasis:
        xs = []
        for i in range(0,DIM):
            x1D, w1D = poly[i].GaussQuadrature(Ns[i])
            xs.append(x1D)
        V = pp.GradVandermonde(xs, Ns, np.zeros(DIM,dtype=int), None)
    
    # Create Dictionary
    if allocBasis:
        dictionary = {'x': x,
                      'w': w,
                      'nrmw': nrmw,
                      'vals': vals,
                      'V': V}
    else:
        dictionary = {'x': x,
                      'w': w,
                      'nrmw': nrmw,
                      'vals': vals}
    return dictionary

def UQ_PCM_Tensor(solve,params,solvePostProcessing,dists,paramUpdate,polys,distsPoly,Ns,method):
    """
    DESCRIPTION:
        UQ_PCM_Tensor(solve,params,solvePostProcessing,dists,paramUpdate,polys,distsPoly,N,method)
        Performs Uncertainty Quantification analysis of the system described by rhs using the Probabilistic Collocation Method by Tensor Product of the random space.
    INPUT:
        - solve: function handle to the experiment to be quantified. Signature: sol = solve(sample,params)
        - params: parameters to be passed to the ODE system (these contain the uncertain parameter as well)
        - solvePostProcessing: post processing function called after all the simulation have been run. Signature: sols = solvePostProcessing(sols)
        - dists: list of n scipy.stats distribution functions for each uncertain parameter
        - paramUpdate: function handle for updating the set of uncertain parameters. Signature: params = paramUpdate(params,sample)
        - polys: list of n Spectral1D polynomials using for approximating the distribution of the random parameters
        - distsPoly: list of n standard distributions associated to the selected approximating polynomials
        - Ns: list of n integers for the order of polynomial expansion in each random direction
        - method: 'projection' use projection rule in order to compute statistics, 'cubature' use cubature rule in order to compute statistics
    OUTPUT: dictionary with the following attributes
        - 'Samples': list of collocation points (samples) used for UQ
        - 'Y': solutions at collocation points
        - 'YHat': Galerkin coefficients (available only if method == 'projection')
        - 'meanSol': estimated mean
        - 'varSol': estimated variance
    """    

    from mpi4py import MPI
        
    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()
    
    if (len(dists) != len(distsPoly) or len(distsPoly) != len(polys) or len(polys) != len(Ns)) :
        print("The number of elements in dist, distPoly and poly is not consistent")
        return
    else:
        DIM = len(dists)

    if (method != 'projection') and (method != 'cubature'):
        print("The selected method '%s' does not exist." % (method))
        return
    
    ###################################################################################
    #                              PRE - PROCESSING                                   #
    ###################################################################################
    
    if method == 'projection':
        dictionary = PCM_Tensor(dists,polys,distsPoly,Ns,allocBasis=True)
        V = dictionary['V']
    if method == 'cubature':
        dictionary = PCM_Tensor(dists,polys,distsPoly,Ns,allocBasis=False)
    w = dictionary['w']
    vals = dictionary['vals']
    Ncoll = len(vals)
    
    ###################################################################################
    #                              SIMULATION                                         #
    ###################################################################################
        
    def action(sols,sol):
        sols.append(sol)
        return sols
        
    solutions = run_experiments(solve,vals,params,paramUpdate,action)
    
    if myrank == 0:
        solutions = solvePostProcessing(solutions)    
    
    ###################################################################################
    #                              POST-PROCESSING                                    #
    ###################################################################################
    
    if myrank == 0:
        # Reshape the solutions so that they are [Ncoll x NDOF], where NDOF is the total DOF
        # Example: for a time dep. ode NDOF = DOF x len(t)
        solsShape = solutions[0].shape
        NDOF = np.prod(solsShape)
        solsRS = np.zeros( (Ncoll, NDOF) )
        for i in range(0,Ncoll):
            # For each solution
            solsRS[i,:] = np.reshape(solutions[i],np.prod(solsShape))
        
        # Post processing for the computation of the PCM coefficients
        # using discrete projection
        def PCM_PostProcProjection(sols):
            # Compute gPC post processing on sols
            # with the following order: DOF, samples, sols
            solsHat = np.zeros(sols.shape)
            A = V * np.tile(w,(1,V.shape[1]))
            for i in range(0,NDOF):
                # For each NDOF
                solsHat[:,i] = np.dot(A.T,sols[:,i])
            return solsHat
        
        if method == 'projection':
            solsHatRS = PCM_PostProcProjection(solsRS)
            
            # Compute mean
            gamma = polys[0].Gamma(0)
            for i in range(1,DIM):
                gamma = gamma * polys[i].Gamma(0)
            meanSol = solsHatRS[0,:] / np.sqrt(gamma)
            
            # Compute variance
            varSol = np.zeros(NDOF)
            for i in range(0,NDOF):
                    varSol[i] = sum(solsHatRS[1:,i]**2.)
        
        if method == 'cubature':
            meanSol = np.zeros(NDOF)
            varSol = np.zeros(NDOF)
            
            for i in range(0,NDOF):
                meanSol[i] = np.dot(solsRS[:,i],w)
                varSol[i] = np.dot(solsRS[:,i]**2.,w) - meanSol[i]**2.
        
        # Reshape the solutions and the statistics to the original shapes
        sols = np.zeros((Ncoll,)+solsShape)
        for i in range(0,Ncoll):
            sols[i,:] = np.reshape(solsRS[i,:],solsShape)
        meanSol = np.reshape(meanSol,solsShape)
        varSol = np.reshape(varSol,solsShape)
        if method == 'projection':
            solsHat = np.zeros((Ncoll,)+solsShape)
            for i in range(0,Ncoll):
                solsHat[i,:] = np.reshape(solsHatRS[i,:],solsShape)
            
        ###################################################################################
        #                              OUTPUT                                             #
        ###################################################################################
        # Prepare a dictionary with necessary data
        if method == 'projection':
            dictionary = {'Samples': vals,
                          'Y': sols,
                          'YHat': solsHat,
                          'meanSol': meanSol,
                          'varSol': varSol}
        if method == 'cubature':
            dictionary = {'Samples': vals,
                          'Y': sols,
                          'meanSol': meanSol,
                          'varSol': varSol}
        return dictionary

def UQ_SparseGrid(solve,params,solvePostProcessing,dists,paramUpdate,method,k,sym=1):
    """
    DESCRIPTION:
        UQ_PCM_Tensor(solve,params,solvePostProcessing,dists,paramUpdate,polys,distsPoly,N,method)
        Performs Uncertainty Quantification analysis of the system described by rhs using the Probabilistic Collocation Method by Tensor Product of the random space.
    INPUT:
        - solve: function handle to the experiment to be quantified. Signature: sol = solve(sample,params)
        - params: parameters to be passed to the ODE system (these contain the uncertain parameter as well)
        - solvePostProcessing: post processing function called after all the simulation have been run. Signature: sols = solvePostProcessing(sols)
        - dists: list of n scipy.stats distribution functions for each uncertain parameter
        - paramUpdate: function handle for updating the set of uncertain parameters. Signature: params = paramUpdate(params,sample)
        - method: Sparse Grid method selected. See description below.
        - k: accuracy of the Sparse Grid rule
        - sym: [optional, default = 1] symmetry parameter for Sparse Grid package
    OUTPUT: dictionary with the following attributes
        - 'Samples': list of collocation points (samples) used for UQ
        - 'Y': solutions at collocation points
        - 'meanSol': estimated mean
        - 'varSol': estimated variance
    Description:
        The methods available for sparse grid are:
            * KPU = Nested rule for unweighted integral over [0,1]
            * KPN = Nested rule for integral with Gaussian weight
            * GQU = Gaussian quadrature for unweighted integral over [0,1] (Gauss-Legendre)
            * GQN = Gaussian quadrature for integral with Gaussian weight (Gauss-Hermite)
            * func =  any function provided by the user that accept level l and returns nodes n and weights w for univariate quadrature rule with polynomial exactness 2l-1 as [n w] = feval(func,level)
    """    

    from mpi4py import MPI
    
    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()
    
    DIM = len(dists)

    if (method != SparseGrids.KPU) and (method != SparseGrids.KPN) and (method != SparseGrids.GQU) and (method != SparseGrids.GQN) and (not callable(method)):
        print("The selected method '%s' does not exist." % (method))
        return
    
    ###################################################################################
    #                              PRE - PROCESSING                                   #
    ###################################################################################
    
    sg = SparseGrids.SparseGrid(method, DIM, k, sym);
    (n,w) = sg.sparseGrid();
    
    vals = n.copy();
    if (method == SparseGrids.KPN) or (method == SparseGrids.GQN):
        ndist = stats.norm(0.,1.);
        for i in range(0,DIM):
            vals[:,i] = dists[i].ppf(ndist.cdf(n[:,i]))
    elif (method == SparseGrids.KPU) or (method == SparseGrids.GQU):
        udist = stats.uniform(0.,1.);
        for i in range(0,DIM):
            vals[:,i] = dists[i].ppf(udist.cdf(n[:,i]))
    
    Ncoll = vals.shape[0]
    
    ###################################################################################
    #                              SIMULATION                                         #
    ###################################################################################
        
    def action(sols,sol):
        sols.append(sol)
        return sols
        
    solutions = run_experiments(solve,vals,params,paramUpdate,action)
    
    if myrank == 0:
        solutions = solvePostProcessing(solutions)    
    
    ###################################################################################
    #                              POST-PROCESSING                                    #
    ###################################################################################
    
    if myrank == 0:
        # Reshape the solutions so that they are [Ncoll x NDOF], where NDOF is the total DOF
        # Example: for a time dep. ode NDOF = DOF x len(t)
        solsShape = solutions[0].shape
        NDOF = np.prod(solsShape)
        solsRS = np.zeros( (Ncoll, NDOF) )
        for i in range(0,Ncoll):
            # For each solution
            solsRS[i,:] = np.reshape(solutions[i],np.prod(solsShape))
        
        if method == 'cubature':
            meanSol = np.zeros(NDOF)
            varSol = np.zeros(NDOF)
            
            for i in range(0,NDOF):
                meanSol[i] = np.dot(solsRS[:,i],w)
                varSol[i] = np.dot(solsRS[:,i]**2.,w) - meanSol[i]**2.
        
        # Reshape the solutions and the statistics to the original shapes
        sols = np.zeros((Ncoll,)+solsShape)
        for i in range(0,Ncoll):
            sols[i,:] = np.reshape(solsRS[i,:],solsShape)
        meanSol = np.reshape(meanSol,solsShape)
        varSol = np.reshape(varSol,solsShape)
            
        ###################################################################################
        #                              OUTPUT                                             #
        ###################################################################################
        # Prepare a dictionary with necessary data
        dictionary = {'Samples': vals,
                      'Y': sols,
                      'meanSol': meanSol,
                      'varSol': varSol}
        return dictionary

class MEPCM_Element:
    ###########
    # PARAMETERS CONSTRUCTED BY THE USER OR REFINEMENT PROCEDURE
    DIM = 0;
    dists = []; # Stores the distributions along the D dimensions
    distsCond = []; # Stores the distribution scalings along the D dimensions
    el = []; # Stores the element coordinates along the D dimensions
    Ns = []; # Stores the order along the D dimensions
    finite = False; # True = the element is finite in all the dimensions, False = the element is infinite in at least one dimension
    ###########
    # DIRTY PARAMETER
    dirty = True; # True = tensor grid needs to be generated, False = nothing needs to be done
    ###########
    # PARAMETERS CONSTRUCED BY THE TENSOR GRID GENERATOR
    # 1D Parameters
    poly1D = []; # Stores the polynomial parameters along the D dimensions
    x1D = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
    vals1D = []; # Stores the collocation points needed by the element
    w1D = []; # Stores the weights corresponding to the collocation points
    pr1D = []; # Stores the probability of the element Pr(I_B_k = 1)
    dim1D = []; # Stores the 1D element size
    V1D = []; # Stores the Vandermonde matrix for the element
    finite1D = []; # True = the element is finite the dimension, False = the element is infinite the dimension
    # ND Parameters
    xEl = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
    valsEl = []; # Stores the collocation points needed by the element
    wEl = []; # Stores the weights corresponding to the collocation points
    prEl = 0.0; # Stores the probability of the element Pr(I_B_k = 1)
    dimEl = 0.0; # Stores the N-D hyper-volume of the element
    VEl = []; # Stores the Vandermonde matrix for the element
    orderEl = []; # Stores the tuples containing the order combination for each basis function
    
    def __init__(self, el, dists, distsCond, Ns):
        if (len(el) != len(dists) or len(dists) != len(Ns) or len(Ns) != len(distsCond)) :
            print("The number of elements in rs, Ns, ks and norms is not consistent")
            return
        
        self.DIM = len(el);
        
        self.dists = dists;
        self.distsCond = distsCond;
        self.el = el;
        self.Ns = Ns;
        
        self.poly1D = []; # Stores the polynomial parameters along the D dimensions
        self.x1D = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.vals1D = []; # Stores the collocation points needed by the element
        self.w1D = []; # Stores the weights corresponding to the collocation points
        self.pr1D = []; # Stores the probability of the element Pr(I_B_k = 1)
        self.dim1D = []; # Stores the 1D element size
        self.V1D = []; # Stores the Vandermonde matrix for the element
        self.finite1D = []; # True = the element is finite the dimension, False = the element is infinite the dimension
        # ND Parameters
        self.xEl = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.valsEl = []; # Stores the collocation points needed by the element
        self.wEl = []; # Stores the weights corresponding to the collocation points
        self.prEl = 0.0; # Stores the probability of the element Pr(I_B_k = 1)
        self.dimEl = 0.0; # Stores the N-D hyper-volume of the element
        self.VEl = []; # Stores the Vandermonde matrix for the element
        self.orderEl = []; # Stores the tuples containing the order combination for each basis function
        
        self.prEl = 1.0;
        self.dimEl = 1.0;
        for eld,distd in zip(self.el,self.dists):
            # Compute the elements probability
            pr = distd.cdf(eld[1])-distd.cdf(eld[0]);
            d = eld[1]-eld[0];
            self.pr1D.append(pr);
            self.dim1D.append(d);
            self.prEl *= pr;
            self.dimEl *= d;
        
        # Check if finite element
        self.finite = True;
        for eld in self.el:
            self.finite1D.append(not any(abs(eld) == np.inf));
            self.finite = self.finite and not any(abs(eld) == np.inf);
        
        self.dirty = True;
    
    def __getstate__(self):
        # Store parameters for polynomials
        poly1D_params = []
        for pp in self.poly1D:
            # Empty space for the measure (it is reconstructed at __setstate__ time)
            params = list(pp.params)
            params[5] = None
            poly1D_params.append( params )
        
        return dict(DIM = self.DIM,
#                    dists = self.dists,
#                    distsCond = self.distsCond,
                    el = self.el,
                    Ns = self.Ns,
                    finite = self.finite,
                    dirty = self.dirty,
                    poly1D_params = poly1D_params,
                    x1D = self.x1D,
                    vals1D = self.vals1D,
                    w1D = self.w1D,
                    pr1D = self.pr1D,
                    dim1D = self.dim1D,
                    V1D = self.V1D,
                    finite1D = self.finite1D,
                    xEl = self.xEl,
                    valsEl = self.valsEl,
                    wEl = self.wEl,
                    prEl = self.prEl,
                    dimEl = self.dimEl,
                    VEl = self.VEl,
                    orderEl = self.orderEl);
    
#    def __reduce__(self):
#        return (MEPCM_Element, ([],[],[],[]), self.__getstate__());
    
    def __setstate__(self,state,dists=None,distsCond=None):
        ###########
        # PARAMETERS CONSTRUCTED BY THE USER OR REFINEMENT PROCEDURE
        self.DIM = state['DIM'];
#        self.dists = state['dists']; # Stores the distributions along the D dimensions
#        self.distsCond = state['distsCond']; # Stores the distribution scalings along the D dimensions
        self.el = state['el']; # Stores the element coordinates along the D dimensions
        self.Ns = state['Ns']; # Stores the order along the D dimensions
        self.dists = dists;
        self.distsCond = distsCond;
        self.finite = state['finite']; # True = the element is finite in all the dimensions, False = the element is infinite in at least one dimension
        ###########
        # DIRTY PARAMETER
        self.dirty = state['dirty']; # True = tensor grid needs to be generated, False = nothing needs to be done
        ###########
        # PARAMETERS CONSTRUCED BY THE TENSOR GRID GENERATOR
        # 1D Parameters
        self.poly1D = []
        for d,params in enumerate(state['poly1D_params']):
            # Restore measure function definition
            if (abs(self.el[d][0]) == np.inf) or (abs(self.el[d][1]) == np.inf):
                # No mapping to [-1,1]. TAILS ARE ASSUMED TO GET RAPIDLY TO 0
                def wf_cap(y,i):
                    return self.dists[d].pdf(self.distCond[d](y,-1))/self.prEl;
            else:
                def wf_cap(y,i):
                    centEl = (self.el[d][1] + self.el[d][0])/2.;
                    dimEl = (self.el[d][1]-self.el[d][0]);
                    return self.dists[d].pdf(y * dimEl/2. + centEl)/self.prEl * dimEl * 0.5**self.DIM;
            
            params[5] = wf_cap
            self.poly1D.append(Spectral1D.Poly1D(Spectral1D.ORTHPOL,params))
#        self.poly1D = state['poly1D']; # Stores the polynomial parameters along the D dimensions
        self.x1D = state['x1D']; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.vals1D = state['vals1D']; # Stores the collocation points needed by the element
        self.w1D = state['w1D']; # Stores the weights corresponding to the collocation points
        self.pr1D = state['pr1D']; # Stores the probability of the element Pr(I_B_k = 1)
        self.dim1D = state['dim1D']; # Stores the 1D element size
        self.V1D = state['V1D']; # Stores the Vandermonde matrix for the element
        self.finite1D = state['finite1D']; # True = the element is finite the dimension, False = the element is infinite the dimension
        # ND Parameters
        self.xEl = state['xEl']; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.valsEl = state['valsEl']; # Stores the collocation points needed by the element
        self.wEl = state['wEl']; # Stores the weights corresponding to the collocation points
        self.prEl = state['prEl']; # Stores the probability of the element Pr(I_B_k = 1)
        self.dimEl = state['dimEl']; # Stores the N-D hyper-volume of the element
        self.VEl = state['VEl']; # Stores the Vandermonde matrix for the element
        self.orderEl = state['orderEl']; # Stores the tuples containing the order combination for each basis function
    
    def __clear1D(self):
        self.poly1D = []; # Stores the polynomial parameters along the D dimensions
        self.x1D = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.vals1D = []; # Stores the collocation points needed by the element
        self.w1D = []; # Stores the weights corresponding to the collocation points
        self.V1D = []; # Stores the Vandermonde matrix for the element
    
    def __clearTensor(self):
        self.xEl = []; # Stores the collocation points needed by the element (in the scaled element reference frame)
        self.valsEl = []; # Stores the collocation points needed by the element
        self.wEl = []; # Stores the weights corresponding to the collocation points
        self.VEl = []; # Stores the Vandermonde matrix for the element
        self.orderEl = []; # Stores the tuples containing the order combination for each basis function
    
    def getstate(self):
        return self.__getstate__();
    
    def setstate(self,state,dists=None,distsCond=None):
        self.__setstate__(state,dists,distsCond);
    
    def __InitBasis1D(self, allocBasis=True):
        """
        Basis1D([allocBasis=True])
        Generates the 1D basis using the list of distributions of the parameters
        space approximated by properly construced orthogonal polynomials.
        INPUT:
            ``allocBasis``: boolean. True if you want the tensor basis.
        OUTPUT:
        """
        
        self.__clear1D();
        
        #############
        # Generate 1D basis for each dimension
        for d in range(0,self.DIM):
            dist = self.dists[d];
            N = self.Ns[d];
            eld = self.el[d];
            distCond = self.distsCond[d];
            
            # Scale the element using user provided function
            elSC = distCond(eld, 1);
            
            # Prepare the measure function
            if (abs(eld[0]) == np.inf) or (abs(eld[1]) == np.inf):
                # No mapping to [-1,1]. TAILS ARE ASSUMED TO GET RAPIDLY TO 0
                def wf_cap(y,i):
                    return dist.pdf(distCond(y,-1))/self.prEl;
            else:
                def wf_cap(y,i):
                    centEl = (eld[1] + eld[0])/2.;
                    dimEl = (eld[1]-eld[0]);
                    return dist.pdf(y * dimEl/2. + centEl)/self.prEl * dimEl * 0.5**self.DIM;
            
            # Generate basis and Gauss points
            ncapm = 500;
            mc = 1;
            mp = 0;
            xp = np.array([]);
            yp = np.array([]);

            irout = 0; # 1: Use sti internally, 0: Use lancz
            if (abs(elSC[0]) == np.inf) or (abs(elSC[1]) == np.inf):
                # Do not apply scaling (we assume tails decay!!)
                x = elSC;
                if (abs(elSC[0]) == np.inf): finl = False; 
                else: finl = True;
                if (abs(elSC[1]) == np.inf): finr = False; 
                else: finr = True;
            else:
                x = np.linspace(-1.,1.,mc+1);
                centEl = (elSC[1] + elSC[0])/2.;
                dimEl = (elSC[1]-elSC[0])/2.;
                finr = True;
                finl = True;
            endl = x[:-1];
            endr = x[1:];
            
            params = (ncapm, mc, mp, xp, yp, wf_cap, irout,finl,finr,endl,endr)
            P = Spectral1D.Poly1D(Spectral1D.ORTHPOL,params)
            
            self.poly1D.append(P)

            (xg,wg) = P.GaussQuadrature(N)
            wg /= np.sum(wg);
            
            if finl and finr:
                x = xg * dimEl + centEl;
                self.x1D.append(x);
            else:
                x = xg;
                self.x1D.append(x);
            self.w1D.append(wg);
            self.vals1D.append(distCond(x,-1));
            
            if allocBasis:
                VCap = self.Vandermonde1D(self.vals1D[-1],d);
                self.V1D.append(VCap);

    def __InitTensor(self, allocBasis=True):
        """
        Tensor([allocBasis=True])
        Generates the full tensor grid using the list of distributions of the parameters
        space approximated by properly construced orthogonal polynomials.
        INPUT:
            ``allocBasis``: boolean. True if you want the tensor basis.
        OUTPUT:
        """
        
        self.__clearTensor();
        
        #############
        # Generate ND basis
        xs = [];
        vals = [];
        ords = [];
        wKron = np.array([1]);
        if allocBasis:
            VKron = np.array([1]);
        for d in range(0,self.DIM):
            ords.append(range(0,self.Ns[d]))
            xs.append(self.x1D[d]);
            vals.append(self.vals1D[d]);
            wKron = np.kron(wKron, self.w1D[d]);
            if allocBasis:
                VKron = np.kron(VKron, self.V1D[d]);
        self.xEl = np.asarray(list(itertools.product(*xs)))
        self.valsEl = np.asarray(list(itertools.product(*vals)))
        self.orderEl = list(itertools.product(*ords))
        self.wEl = wKron;
        if allocBasis:
            self.VEl = VKron;
    
    def Vandermonde1D(self,x,d):
        if not self.finite1D[d]:
            # Map to the Scaled values
            xSC = self.distsCond[d](x,1);
        if self.finite1D[d]:
            # Scale to [-1,1] if dimension is finite
            centEl = (self.el[d][1] + self.el[d][0])/2.
            dimEl = (self.el[d][1] - self.el[d][0])/2.
            xSC = (x - centEl)/dimEl
        return self.poly1D[d].GradVandermonde1D(xSC,self.Ns[d],0,norm=False);
    
    def TensorVandermonde(self,x,usekron=True):
        """
        Takes a 2d-array (preferibly (N,DIM)) or list (of array of points for each dimension)
        If kron == True : take kron product of 1D vandermonde matrices
        If kron == False : take column multiplication of 1D matrices
        """
        
        if type(x) == list:
            x = np.array(x);
        
        if (x.shape[0] == self.DIM) and (x.shape[1] != self.DIM):
            x = x.T;
        
        if (x.shape[1] != self.DIM) :
            print("The number of elements in x is not consistent with the dimension of the element")
            return
        
        if (x.shape[0] == 0):
            print("The number of points in x is 0.")
            return
        
        if usekron:
            VKron = np.array([1.]);
            for d in range(0,self.DIM):
                VKron = np.kron(VKron, self.Vandermonde1D(x[:,d],d));
        else:
            VKron = np.ones((x.shape[0],1))
            for d in range(0,self.DIM):
                VKronNew = np.zeros((VKron.shape[0],VKron.shape[1] * (self.Ns[d]+1)))
                V = self.Vandermonde1D(x[:,d],d);
                for col in range(0,VKron.shape[1]):
                    VKronNew[:,col*(self.Ns[d]+1):(col+1)*(self.Ns[d]+1)] = np.tile(VKron[:,col],(self.Ns[d]+1,1)).T * V
                VKron = VKronNew
        return VKron;
    
    def GammaPower(self,power):
        maxBasis = self.VEl.shape[1];
        dims = np.ones(power) * maxBasis;
        EIdxs = list(itertools.product(*[range(0,maxBasis) for i in range(0,power)]));
        E = np.zeros(dims);
        # Compute the number of gauss points needed in order to get exact approximation
        Ngauss = [];
        for N in self.Ns:
            Ngauss.append( int(np.ceil( power*N/2. )) );
        
        # Compute recursion coefficients for the number of gammas required
        poly1D = [];
        x1D = [];
        w1D = [];
        vals1D = [];
        V1D = [];
        for d in range(0,self.DIM):
            dist = self.dists[d];
            N = Ngauss[d];
            eld = self.el[d];
            distCond = self.distsCond[d];
            
            # Scale the element using user provided function
            elSC = distCond(eld, 1);
            
            # Prepare the measure function
            if (abs(eld[0]) == np.inf) or (abs(eld[1]) == np.inf):
                # No mapping to [-1,1]. TAILS ARE ASSUMED TO GET RAPIDLY TO 0
                def wf_cap(y,i):
                    return dist.pdf(distCond(y,-1))/self.prEl;
            else:
                def wf_cap(y,i):
                    centEl = (eld[1] + eld[0])/2.;
                    dimEl = (eld[1]-eld[0]);
                    return dist.pdf(y * dimEl/2. + centEl)/self.prEl * dimEl * 0.5**self.DIM;
            
            # Generate basis and Gauss points
            ncapm = 500;
            mc = 1;
            mp = 0;
            xp = np.array([]);
            yp = np.array([]);
            irout = 0; # 1: Use sti internally, 0: Use lancz
            if (abs(elSC[0]) == np.inf) or (abs(elSC[1]) == np.inf):
                # Do not apply scaling (we assume tails decay!!)
                x = elSC;
                if (abs(elSC[0]) == np.inf): finl = False; 
                else: finl = True;
                if (abs(elSC[1]) == np.inf): finr = False; 
                else: finr = True;
            else:
                x = np.linspace(-1.,1.,mc+1);
                centEl = (elSC[1] + elSC[0])/2.;
                dimEl = (elSC[1]-elSC[0])/2.;
                finr = True;
                finl = True;
            endl = x[:-1];
            endr = x[1:];
            
            params = (ncapm, mc, mp, xp, yp, wf_cap, irout,finl,finr,endl,endr)
            P = Spectral1D.Poly1D(Spectral1D.ORTHPOL,params)
            
            self.poly1D.append(P)
            
            (xg,wg) = P.GaussQuadrature(N)
            wg /= np.sum(wg);
            
            if finl and finr:
                x = xg * dimEl + centEl;
                x1D.append(x);
            else:
                x = xg;
                x1D.append(x);
            w1D.append(wg);
            vals1D.append(distCond(x,-1));
            
            VCap = self.Vandermonde1D(vals1D[-1],d);
            V1D.append(VCap);
        
        # Compute tensor product of the Vandermonde matrices and weights
        wKron = np.array([1]);
        VKron = np.array([1]);
        for d in range(0,self.DIM):
            wKron = np.kron(wKron, w1D[d]);
            VKron = np.kron(VKron, V1D[d]);
        
        # Compute recursion coeff
        for idx in EIdxs:
            prod = VKron[:,idx[0]];
            for i in range(1,power):
                prod = prod * VKron[:,idx[i]];
            E[idx] = np.dot(prod,wKron);
        
        return E;
    
    def isInElement(self,x):
        """
        Takes a 2d-array (preferibly (N,DIM)) or list (of array of points for each dimension)
        """
        
        if type(x) == list:
            x = np.array(x);
        
        if (x.shape[0] == self.DIM) and (x.shape[1] != self.DIM):
            x = x.T;
        
        if (x.shape[1] != self.DIM) :
            print("The number of elements in x is not consistent with the dimension of the element")
            return
        
        # Check belonging
        belonging = np.ones(x.shape[0],dtype=bool); # Init all true
        for d in range(0,self.DIM):
            elBelonging = np.all([(x[:,d] >= self.el[d][0]),(x[:,d] < self.el[d][1])],axis=0)
            belonging = np.all([belonging, elBelonging], axis=0)
        return belonging
    
    def refresh(self,allocBasis=True):
        self.__InitBasis1D(allocBasis);
        self.__InitTensor(allocBasis);
        self.dirty = False;

class MEPCM_Forest:
    DIM = 0;
    elements = [];
    dists1D = [];
    distsCond1D = [];
    
    def __init__(self, el1D=None, dists1D=None, distsCond1D=None, Ns1D=None):
        if el1D == None:
            self.elements = [];
            self.DIM = 0;
            return
        
        if (len(el1D) != len(dists1D) or len(dists1D) != len(Ns1D) or len(Ns1D) != len(distsCond1D)) :
            print("The number of elements in el1D, dists1D, distsCond1D is not consistent")
            return

        self.DIM = len(el1D);
        self.dists1D = dists1D;
        self.distsCond1D = distsCond1D;
        
        self.elements = self.split(el1D, Ns1D);
    
#    def __reduce__(self):
#        return (MEPCM_Forest, (None,None,None,None), self.__getstate__());
    
    def __getstate__(self):
        return dict(DIM = self.DIM,
                    elements = [e.getstate() for e in self.elements])
#                    dists1D = self.dists1D,
#                    distsCond1D = self.distsCond1D)
    
    def __setstate__(self,state,dists=None,distsCond=None):
        self.DIM = state['DIM'];
        self.elements = []
        self.dists1D = dists;
        self.distsCond1D = distsCond;
        for e in state['elements']:
            el = MEPCM_Element([],[],[],[]);
            el.setstate(e,dists,distsCond);
            self.elements.append(el);
    
    def getstate(self):
        return self.__getstate__();
    
    def setstate(self,state,dists=None,distsCond=None):
        self.__setstate__(state,dists,distsCond);
    
    def split(self, el1D, Ns1D):
        elsIdx = [];
        for d in range(0,self.DIM):
            elsIdx.append(range(0,len(el1D[d])));
        elsIdxList = list(itertools.product(*elsIdx));
        
        root = [];
        for e in range(0,len(elsIdxList)):
            elsIdx = elsIdxList[e];
            el = []; # Coordinates for the element in each dimension
            Ns = [];
            for d in range(0,self.DIM):
                el.append( el1D[d][elsIdx[d]] );
                Ns.append( Ns1D[d][elsIdx[d]] );
            newEl = MEPCM_Element(el,self.dists1D,self.distsCond1D,Ns);
            newEl.refresh();
            root.append( newEl );
        return root;
    
    def refine(self, toRefIndices, toRefDims, toRefOrders):
        """
            ``refineOrder``: 0 = let the order be equal to the parent element. 1 = drop the order of 1
        """
        refinedIndices = [];
        for toRefIdx, toRefDim, toRefOrder in zip(reversed(toRefIndices),reversed(toRefDims),reversed(toRefOrders)):
            el = self.elements[toRefIdx].el;
            Ns = self.elements[toRefIdx].Ns;
            for d in range(0,self.DIM):
                if toRefDim[d]:
                    eld = el[d];
                    if (abs(eld[0]) == np.inf) or (abs(eld[1]) == np.inf):
                        print("Error: you are trying to refine an infinite element")
                        return
                    dimEld = eld[1]-eld[0];
                    el[d] = [np.array([eld[0],eld[0]+dimEld/2.]),np.array([eld[0]+dimEld/2.,eld[1]])];
                    Ns[d] = [toRefOrder[d],toRefOrder[d]];
                else:
                    el[d] = [el[d]];
                    Ns[d] = [toRefOrder[d]];
            newEls = self.split(el,Ns);
            
            self.elements.pop(toRefIdx);
            for newEl, i in zip(reversed(newEls), range(0,len(newEls))):
                self.elements.insert(toRefIdx,newEl);
                if i > 0:
                    for j in range(0,len(refinedIndices)):
                        refinedIndices[j] = refinedIndices[j] + 1
                refinedIndices.append(toRefIdx);
        return list(reversed(refinedIndices));
 
def UQ_MEPCM_Tensor(domain,solve,params,solvePostProcessing,paramUpdate,refOrder,minRefOrd=2,maxRefOrd=2,iMax=10,gamma=0.5,theta1=1e-2, theta2=.75):
    """
    DESCRIPTION:
        UQ_MEPCM_Tensor(solve,params,solvePostProcessing,dists,paramUpdate,polys,distsPoly,N,method)
        Performs Uncertainty Quantification analysis of the system described by rhs using the Probabilistic Collocation Method by Tensor Product of the random space.
    INPUT:
        ``domain``: (MEPCM_Forest object) initial domain on which to evaluate the functions
        ``solve``: function handle to the experiment to be quantified. Signature: sol = solve(sample,params)
        ``params``: parameters to be passed to the ODE system (these contain the uncertain parameter as well)
        ``solvePostProcessing``: post processing function called after all the simulation have been run. Signature: ``solsEl = solvePostProcessing(solsEl)``. It returns properly aligned solutions per element, meaning that for the solutions in each element solsEl[i] contains an [Ncoll x NDOF] array.
        ``paramUpdate``: function handle for updating the set of uncertain parameters. Signature: params = paramUpdate(params,sample)
        ``refinement``: 0 = No refinement, 1 = refinement + partial limiting, 2 = refinement + final limiting
        ``iMax``: maximum number of refinement iterations
        ``gamma``: parameter for adaptivity refinement criteria
        ``theta1``: parameter for adaptivity refinement criteria
        ``theta2``: parameter for the dimensional refinement criteria
    OUTPUT: dictionary with the following attributes
        ``domain``: MEPCM_Forest object
        ``solsEl``: (list #<multi-dim El>) solutions at collocation points per each element
        ``meanSol``: estimated mean
        ``varSol``: estimated variance
        ``uHatEl``: (list #<multi-dim El>) Galerkin coefficients (available only if method == 'projection')
        ``fEval``: total number of solve evalutations
    """

    from mpi4py import MPI
    
    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()
    
    ###################################################################################
    #                              PRE - PROCESSING                                   #
    ###################################################################################
        
    def action(sols,sol):
        sols.append(sol)
        return sols
    
    iLoop = 0;
    stopCrit = False;
    toRefIndices = [];
    refinedIdx = range(0,len(domain.elements));
    solsEl = [];
    totComputations = 0;
    while not stopCrit and (iLoop<iMax):
        iLoop = iLoop+1
        
        if myrank == 0:
            print("Total number of new elements: %d" % len(refinedIdx))
            table = PrettyTable(['#el','PrEl','refined']);
            for eli in range(0,len(domain.elements)):
                table.add_row([eli, domain.elements[eli].prEl, (eli in refinedIdx)])
            print(table)
        
        ###################################################################################
        #                             LOOP SIMULATION                                     #
        ###################################################################################
        
        # Merge all the collocation points
        vals = np.zeros((0,domain.DIM))
        splitSizes = [];
        for eli in refinedIdx:
            vals = np.vstack((vals,domain.elements[eli].valsEl));
            splitSizes.append(domain.elements[eli].valsEl.shape[0]);
        splitPos = np.cumsum(splitSizes)
        totComputations += vals.shape[0];
        
        # Run the experiments
        sols = run_experiments(solve, vals, params, paramUpdate, action);
        
        if myrank == 0:
            # Split solutions by element
            solsElSplit = np.split(sols,splitPos[:-1],axis=0);
            
            # Insert new sols in solsEl
            for eli,i in zip(refinedIdx,range(0,len(refinedIdx))):
                solsEl.insert(eli,solsElSplit[i]);
        
        solsEl = comm.bcast(solsEl, root=0);
    
        ###################################################################################
        #                             LOOP POST-PROCESSING                                #
        ###################################################################################
        
        if myrank == 0:
            # COMPUTE COEFFICIENTS of expansion
            nEls = len(domain.elements);
            uHat = [];
            sigmaEl = [];
            etaEl = [];
            for i in range(0,nEls):
                uh = la.solve(domain.elements[i].VEl,solsEl[i])
                uHat.append( uh )
                if len(solsEl[i]) > 1:
                    Np = min(domain.elements[i].Ns)-1
                    orderList = np.array([sum(o) for o in domain.elements[i].orderEl])
                    gammas = np.dot(domain.elements[i].wEl,domain.elements[i].VEl*domain.elements[i].VEl)
                    sigma2bar = np.dot(uh[1:]**2., gammas[1:] )
                    sigmaEl.append( sigma2bar )
                    if sigma2bar != 0.:
                        eta = np.dot(uh[orderList >= Np]**2., gammas[orderList >= Np]) / sigma2bar
                        etaEl.append( eta );
                    else: etaEl.append(0.);
                else:
                    sigmaEl.append(0.);
                    etaEl.append(0.);
                
            # Check stopping criteria
            toRefIndices = [];
            toRefDims = [];
            toRefOrders = []; # Order change in each dimension
            stopCrit = True;
            if (iLoop < iMax):
                for eli in reversed(range(0,len(uHat))):
                    if (etaEl[eli]**gamma * domain.elements[eli].prEl >= theta1):
                        toRefIndices.append(eli);
                        # Detect the dimension to be refined
                        dimSens = np.zeros(domain.DIM);
                        orderEl = np.array(domain.elements[eli].orderEl)
                        orderList = np.array([sum(o) for o in domain.elements[eli].orderEl])
                        for d in range(0,domain.DIM):
                            Np = domain.elements[eli].Ns[d]-1
                            NpIdx = (orderEl[:,d] == Np) * (orderList == Np);
                            gammaNp = np.dot(domain.elements[eli].wEl,domain.elements[eli].VEl[:,NpIdx]*domain.elements[eli].VEl[:,NpIdx])
                            dimSens[d] = (uHat[eli][NpIdx][0] * gammaNp[0])/ (etaEl[eli] * sigmaEl[eli]);
                        
                        maxSens = np.amax(np.abs(dimSens));
                        refDims = np.zeros(domain.DIM,dtype=bool);
                        refOrders = np.zeros(domain.DIM,dtype=int);
                        for d in range(0,domain.DIM):
                            if domain.elements[eli].finite1D[d] and (np.abs(dimSens[d]) >= theta2 * maxSens):
                                refDims[d] = True;
                                refOrders[d] = max(minRefOrd,min(domain.elements[eli].Ns[d]+refOrder,maxRefOrd));
                            else:
                                refDims[d] = False;
                                # Check eta condition for lower coefficients
                                Np = domain.elements[eli].Ns[d]-1; # Remember Ns-1 is the highest order
                                sens = 0.0;
                                while (Np > minRefOrd) and (np.abs(sens) < theta2 * maxSens):
                                    Np -= 1;
                                    NpIdx = (orderEl[:,d] == Np) * (orderList == Np);
                                    gammaNp = np.dot(domain.elements[eli].wEl,domain.elements[eli].VEl[:,NpIdx]*domain.elements[eli].VEl[:,NpIdx])
                                    sens = (uHat[eli][NpIdx][0] * gammaNp[0])/ (etaEl[eli] * sigmaEl[eli]);
                                refOrders[d] = Np+1;
                        
                        toRefDims.append(refDims);
                        toRefOrders.append(refOrders);
                        if (iLoop < iMax): solsEl.pop(eli);
                        stopCrit = False;
            
            toRefIndices = list(reversed(toRefIndices));
            toRefDims = list(reversed(toRefDims));
            toRefOrders = list(reversed(toRefOrders));
            
            table = PrettyTable(['#el','PrEl', 'eta', 'etaPr','toRefine']);
            for eli in range(0,len(domain.elements)):
                table.add_row([eli, domain.elements[eli].prEl, etaEl[eli], etaEl[eli]**gamma * domain.elements[eli].prEl, (eli in toRefIndices)])
            print(table)
            
            if not stopCrit and (iLoop<iMax):
                # Split elements (exclude tails)
                refinedIdx = domain.refine(toRefIndices,toRefDims,toRefOrders);
            
            domainState = domain.getstate()
        
        else:
            # Other processes
            domainState = [];
        
        stopCrit = comm.bcast(stopCrit, root=0);
        domainState = comm.bcast(domainState, root=0);
        refinedIdx = comm.bcast(refinedIdx, root=0);
        domain.setstate(domainState,domain.dists1D,domain.distsCond1D)
            
    ## END LOOPS ##
    
    if myrank == 0:
        print("Total number of experiments: %d" % totComputations)
        
        # COMPUTE MEAN
        uHat0 = np.array([uHat[i][0] for i in range(0,len(uHat))]);
        prEls = np.array([domain.elements[i].prEl for i in range(0,len(domain.elements))]);
        meanSol = np.dot(uHat0, np.asarray(prEls));
        
        # COMPUTE VARIANCE
        varSol = np.dot( np.array(sigmaEl) + (uHat0 - meanSol)**2. , prEls)
        
        ###################################################################################
        #                              OUTPUT                                             #
        ###################################################################################
        # Prepare a dictionary with necessary data
        dictionary = {'domain': domain,
                      'solsEl': solsEl,
                      'meanSol': meanSol,
                      'varSol': varSol,
                      'uHatEl': uHat,
                      'fEval': totComputations}
        return dictionary



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

import sys
import numpy as np
import numpy.linalg as npla
import scipy.stats as stats
try:
    import cPickle as pkl
except ImportError:
    import pickle as pkl
from SpectralToolbox import Spectral1D
from UQToolbox import CutANOVA
import UQToolbox.RandomSampling as RS
from UQToolbox.unittests.auxiliary import bcolors

def print_ok(string):
    print(bcolors.OKGREEN + "[SUCCESS] " + string + bcolors.ENDC)

def print_fail(string,msg=''):
    print(bcolors.FAIL + "[FAILED] " + string + bcolors.ENDC)
    if msg != '':
        print(bcolors.FAIL + msg + bcolors.ENDC)

def run(maxprocs=None, PLOTTING=True):
    if PLOTTING:
        import matplotlib.pyplot as plt

    store_freq = 100
    out_file = 'sensitivity.pkl'
        
    DIM = 8                                           # number of uncertainties
    dists = [stats.uniform(0,1) for i in range(DIM)] # Distributions

    params = dict( a = np.asarray([0., 1.,4.5, 9., 99., 99., 99., 99.]) )
    def fun(X,params):
        a = params['a']
        if len(X.shape) == 1:
            Y = np.prod( (np.abs(4.*X - 2.) + a)/(1. + a) );
        elif len(X.shape) == 2:
            Y = np.prod( (np.abs(4.*X - 2.) + a)/(1. + a) , 1);
        return Y

    poly_order = 6  # Polynomial chaos order (MUST BE EVEN)
    cut_order = 2
    
    # Polynomial construction for the distribution at hand
    pp = Spectral1D.generate(Spectral1D.JACOBI,[0.,0.])
    polys = [pp for i in range(DIM)]
    dists_polys = [ stats.uniform(-1,2) ] * DIM # orthogonality distribution for JACOBBI(0,0) polynomials (used for scaling).
    orders = [poly_order for i in range(DIM)] # Isotropic expansion (equal for all params)
    X_cut = np.zeros((1,len(polys))) # Anchor point (usually the central one)
    tol = 5. * np.spacing(1) # Internal tolerance for the construction of the nested grids

    cut_HDMR = CutANOVA.CutHDMR( fun, params,
                                 dists, dists_polys, polys,
                                 orders, cut_order, X_cut, tol,
                                 marshal_f = False,
                                 store_file = out_file )
    print_ok("UQToolbox.CutANOVA: CutHDMR construction")

    # Evaluate f on the cutHDMR grid
    print("Evaluation: %d/%d" % (len(cut_HDMR.get_results()), len(cut_HDMR.get_results())+len(cut_HDMR.get_new_samples())))

    cut_HDMR.run( maxprocs, store_freq )
    print_ok("UQToolbox.CutANOVA: CutHDMR evaluation")

    print("Start HDMR computation - cut-HDMR")
    # Compute the cutHDMR
    cut_HDMR.computeCutHDMR()

    print("Start HDMR computation - ANOVA-HDMR")
    # Compute the ANOVA-HDMR
    cut_HDMR.computeANOVA_HDMR()
    print("End HDMR computation")

    print_ok("UQToolbox.CutANOVA: CutHDMR computation")

    # Compute an estimate for the total variance (using Latin Hyper Cube)
    exp_lhc = RS.Experiments(fun,params,dists,False)
    exp_lhc.sample(2000, method='lhc')
    exp_lhc.run()
    MC_exp = np.asarray(exp_lhc.get_results())
    MC_mean = np.mean(MC_exp)
    MC_var = np.var(MC_exp)
    if PLOTTING:
        plt.figure()
        plt.subplot(2,1,1)
        plt.plot(np.array([np.mean(MC_exp[:i]) for i in range(len(MC_exp))]))
        plt.subplot(2,1,2)
        plt.plot(np.array([np.var(MC_exp[:i]) for i in range(len(MC_exp))]))
    print_ok("UQToolbox.CutANOVA: Total variance evaluation")

    # Compute individual variances
    D = []
    for level_grids in cut_HDMR.grids:
        D_level = []
        for grid in level_grids:
            D_level.append( np.dot(grid.ANOVA_HDMR_vals**2., grid.WF) )
        D.append(D_level)
    Var_ANOVA = np.sum( [ np.sum(sd) for sd in D[1:] ] )
    print("Variance ANOVA: %f" % Var_ANOVA)
    print("VarANOVA/TotVar = %f" % (Var_ANOVA/MC_var))

    # Compute Total variances per component
    TV = np.zeros(DIM)
    for idx in range(DIM):
        for level, level_idxs in enumerate(cut_HDMR.idxs):
            for j, idxs in enumerate(level_idxs):
                if idx in idxs:
                    TV[idx] += D[level][j]
    TS = TV/MC_var
    print("N     TV       TS")
    for i,grid in enumerate(cut_HDMR.grids[1]):
        print("%2d    %.4f   %.4f" % (i+1, TV[i], TS[i]))

    print_ok("UQToolbox.CutANOVA: TSI calculation")

    if PLOTTING:
        plt.figure()
        plt.pie(TS/np.sum(TS),labels=["x%d"%(i+1) for i in range(DIM)])
        plt.show(False)


if __name__ == "__main__":
    # Number of processors to be used, defined as an additional arguement 
    # $ python TestKL.py N
    if len(sys.argv) == 2:
        maxprocs = int(sys.argv[1])
    else:
        maxprocs = None

    run(maxprocs)



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#


"""
 Cut ANOVA
=========================================

Created on Wed Mar 13 09:03:41 2013

@author: Daniele Bigoni (dabi@dtu.dk)

Description
-----------

This module is used to construct High Dimensionar Model Representation using cut-HDMR based on spectral methods from the module :py:mod:`SpectralToolbox.Spectral1D`. Additionally the cut-HDMR can be used to compute the associated ANOVA-HDMR.

"""

import numpy as np

import itertools

from SpectralToolbox import SpectralND
from SpectralToolbox import Misc
from UQToolbox import Driver

class NestedGrid:
    
    def __init__(self,parents,XF,WF,V,X_DIM,tol):
        self.parents = []
        self.XF = None
        self.WF = None
        self.V = None       # Vandermonde matrix
        self.X_DIM = None
        self.X_DIM_NOT_OVER = None
        self.tol = None
        
        self.FX_NOT_OVER = None
        
        self.parents = parents  # List of Nested Sparse Grids that are parent to this
        self.XF = XF
        self.WF = WF
        self.V = V
        self.X_DIM = X_DIM
        self.tol = tol
        self.X_DIM_NOT_OVER = self.X_DIM[ np.logical_not(self.get_overlapping()), : ]
    
    def __getstate__(self):
        return dict(parents = self.parents,
                    XF = self.XF,
                    WF = self.WF,
                    V = self.V,
                    X_DIM = self.X_DIM,
                    X_DIM_NOT_OVER = self.X_DIM_NOT_OVER,
                    tol = self.tol,
                    FX_NOT_OVER = self.FX_NOT_OVER)
    
    def __setstate__(self,state):
        self.parents = state['parents']
        self.XF = state['XF']
        self.WF = state['WF']
        self.V = state['V']
        self.X_DIM = state['X_DIM']
        self.X_DIM_NOT_OVER = state['X_DIM_NOT_OVER']
        self.tol = state['tol']
        self.FX_NOT_OVER = state['FX_NOT_OVER']
    
    def get_overlapping(self):
        out = np.zeros(self.XF.shape[0],dtype=bool)
        for nsg in self.parents:
            out = np.logical_or( out, Misc.findOverlapping(self.X_DIM, nsg.X_DIM, self.tol) )
        return out
    
    def get_FX(self):
        FX = np.zeros(self.X_DIM.shape[0])
        overlapping = np.zeros(self.XF.shape[0],dtype=bool)
        for nsg in self.parents:
            over = Misc.findOverlapping(self.X_DIM, nsg.X_DIM, self.tol)
            FX[over] = nsg.get_FX();
            overlapping = np.logical_or( overlapping, over )
        FX[np.logical_not(overlapping)] = self.FX_NOT_OVER
        return FX


class CutHDMRGrid(NestedGrid):
    
    cutHDMR_vals = None
    ANOVA_HDMR_vals = None
    
    def __init__(self,parents,XF,WF,V,X_DIM,tol):
        NestedGrid.__init__(self,parents,XF,WF,V,X_DIM,tol)
        self.cutHDMR_vals = np.zeros(X_DIM.shape[0])
        self.ANOVA_HDMR_vals = np.zeros(X_DIM.shape[0])
    
    def __getstate__(self):
        dd = NestedGrid.__getstate__(self)
        dd.update(cutHDMR_vals = self.cutHDMR_vals, 
                  ANOVA_HDMR_vals = self.ANOVA_HDMR_vals)
        return dd
    
    def __setstate__(self,state):
        NestedGrid.__setstate__(self,state)
        self.cutHDMR_vals = state['cutHDMR_vals']
        self.ANOVA_HDMR_vals = state['ANOVA_HDMR_vals']

class CutHDMR(Driver):
    
    grids = None
    idxs = None
    tol = None
    
    def __init__(self,f,params, dists,dists_polys,polys,
                 Ns,cut_order,X_cut,tol=1e-11,marshal_f=True, store_file=""):

        super(CutHDMR,self).__init__( f, params, marshal_f, store_file )
        
        self.tol = tol
        
        DIM = len(polys)
        idxs_range = range(DIM)
        self.grids = []    # List of sparse grids
        self.idxs = []   # List of indices

        self.serialize_list.extend( ['grids','idxs','tol'] )
        
        # Insert the zero term on the cut
        XF = X_cut.copy()
        WF = np.ones(1)
        V = np.ones(1)
        X_DIM = X_cut.copy()
        self.idxs.append(list(itertools.combinations(idxs_range,0)))
        self.grids.append( [CutHDMRGrid([],XF,WF,V,X_DIM,self.tol)]  )
        
        for order in range(1,cut_order+1):
            
            # Compute index lists
            idxs = list(itertools.combinations(idxs_range,order))
            self.idxs.append(idxs)
            
            SUB_GS_LIST = []
            for i, idx in enumerate(idxs):
                # order here is intended as integration dimension (so order 0 is integration on a point, order 1 is along a line, etc.)
                # Construct sparse grid of level ORD and dimension 'order'
                poly_list = [polys[j] for j in idx]
                Ns_list = [Ns[j] for j in idx]
                pND = SpectralND.PolyND(poly_list)
                (XF,WF) = pND.GaussQuadrature(Ns_list, norm=True)
                V = pND.GradVandermonde(XF,Ns_list,[0 for j in idx],usekron=False)
                
                # Plug the sparse grid in the correct cut in the DIM dimensional space
                X_DIM = np.tile(X_cut.copy(),(XF.shape[0],1))
                X_DIM[:,np.asarray(idx)] = XF
                
                # Find parent sparse grids of order-1
                idxs_parents = list(itertools.combinations(idx,order-1))
                parents = []
                for idx_parent in idxs_parents:
                    pointer_to_parent = self.idxs[order-1].index(idx_parent)
                    parents.append(self.grids[order-1][pointer_to_parent])
                
                SUB_GS_LIST.append( CutHDMRGrid(parents, XF, WF, V, X_DIM, self.tol) )
            
            self.grids.append(SUB_GS_LIST)

        # Get samples into the Driver class (for parallelization)
        for sub_sgs_list in self.grids:
            for nest_sg in sub_sgs_list:
                # map sample from dists_polys to dists
                samples = np.array([ dist.ppf(
                    dist_poly.cdf(
                        nest_sg.X_DIM_NOT_OVER[:,i_d] ) ) for
                        i_d, (dist,dist_poly) in
                        enumerate(zip(dists,dists_polys)) ]).T
                self.new_samples.extend( list( samples ) )
    
    def __getstate__(self):
        return super(CutHDMR,self).__getstate__()
    
    def __setstate__(self,state):
        super(CutHDMR,self).__setstate__( state )
    
    def get_grids(self):
        return self.grids
    
    def get_idxs(self):
        return self.idxs

    def run(self, maxprocs=None, store_freq=None):
        """
        Extends the run function of Driver, reassigning values on the nested grids
        """
        super(CutHDMR,self).run( maxprocs, store_freq )

        # Redistribute results
        i = 0
        for sub_sgs_list in self.grids:
            for nest_sg in sub_sgs_list:
                nest_sg.FX_NOT_OVER = self.results[i:i+nest_sg.X_DIM_NOT_OVER.shape[0]]
                i += nest_sg.X_DIM_NOT_OVER.shape[0]
        
    def evaluateFun(self, f, transformFunc):
        for sub_sgs_list in self.grids:
            for nest_sg in sub_sgs_list:
                nest_sg.FX_NOT_OVER = f(transformFunc(nest_sg.X_DIM_NOT_OVER))
    
    def computeCutHDMR(self):
        """
        Computes the terms f0, f_i, f_ij, etc. and assign them to the ``CutHDMRGrid``s.
        """
        
        # Start from level zero and go up.
        for order_grids,order_idxs in zip(self.grids,self.idxs):
            for level_grid, level_idxs in zip(order_grids,order_idxs):
                # level_grid: grid on which we are computing the CutHDMR
                # level_idxs: indexs associated with the level_grid
                
                # HDMR projection: f^ij(x_i,x_j) - P_i f(x) - P_j f(x) - P_0 f(x)
                level_grid.cutHDMR_vals = level_grid.get_FX()
                # Find all the subsets of level_idxs and subtract to the cutHDMR_vals
                for idxs in Misc.powerset(level_idxs):
                    if idxs != level_idxs:
                        if idxs == ():
                            subgrid = self.grids[0][0]
                            level_grid.cutHDMR_vals -= subgrid.cutHDMR_vals
                        else:
                            # Lookup for the subgrid
                            order_subgrid = len(idxs)
                            idx_subgrid = self.idxs[order_subgrid].index(idxs)
                            subgrid = self.grids[order_subgrid][idx_subgrid]
                            
                            # Find where to subtract the values
                            # Look for the values with the same idxs coordinates
                            level_grid_Xfilt = level_grid.X_DIM[:,list(idxs)]
                            subgrid_Xfilt = subgrid.X_DIM[:,list(idxs)]
                            for X,val in zip(subgrid_Xfilt,subgrid.cutHDMR_vals):
                                isEqual = Misc.almostEqualList(level_grid_Xfilt,X,self.tol)
                                level_grid.cutHDMR_vals[isEqual] -= val
        
    def computeANOVA_HDMR(self):
        """
        Computes the terms f_0, f_i, f_ij, etc. and assign them to :py:data:`CutHDMRGrid.ANOVA_HDMR_vals`.
        """
        
        # Start from level zero and go up
        for order_grids,order_idxs in zip(self.grids,self.idxs):
            for level_grid, level_idxs in zip(order_grids,order_idxs):
                # level_grid: grid on which we are computing the ANOVA-HDMR
                # level_idxs: indexs associated with the level_grid
                
                # HDMR projection: \int_K^{n-l} f^cut(x) - P_i f(x) - P_j f(x) - P_0 f(x)
                # Compute the integral: iterate over the cuts and project onto the level_grid
                for inner_order_grids, inner_order_idxs in zip(self.grids,self.idxs):
                    for inner_level_grid, inner_level_idxs in zip(inner_order_grids,inner_order_idxs):
                        if level_idxs == ():
                            level_grid.ANOVA_HDMR_vals[0] += np.dot(inner_level_grid.cutHDMR_vals,inner_level_grid.WF)
                        else:
                            # project over the intersection betwenn level_idxs and order_idxs
                            inner_proj_idxs = set(inner_level_idxs).difference(set(level_idxs))
                            inner_cut_idxs = set(level_idxs).difference(inner_proj_idxs)                            
                            level_cut_idxs = set(level_idxs).intersection(inner_level_idxs)
                            
                            inner_proj_coord = inner_level_grid.X_DIM[:,np.asarray(list(inner_cut_idxs),dtype=int)]
                            level_proj_coord = level_grid.X_DIM[:,np.asarray(list(level_cut_idxs),dtype=int)]

                            (inner_proj_cut,inner_cut_idxs_list) = Misc.unique_cuts(inner_proj_coord, self.tol, retIdxs=True)
                            (level_proj_cut,level_cut_idxs_list) = Misc.unique_cuts(level_proj_coord, self.tol, retIdxs=True)
                            
                            # Project
                            for inner_prj_idxs,level_cut_idxs in zip(inner_cut_idxs_list,level_cut_idxs_list):
                                weights = inner_level_grid.WF[inner_prj_idxs]
                                weights /= np.sum(weights)
                                level_grid.ANOVA_HDMR_vals[level_cut_idxs] += np.dot( inner_level_grid.cutHDMR_vals[inner_prj_idxs], weights )
                
                
                # find all subsets of level_idxs and subtract to ANOVA_HDMR_vals
                for idxs in Misc.powerset(level_idxs):
                    if idxs != level_idxs:
                        if idxs == ():
                            subgrid = self.grids[0][0]
                            level_grid.ANOVA_HDMR_vals -= subgrid.ANOVA_HDMR_vals
                        else:
                            # Lookup for the subgrid
                            order_subgrid = len(idxs)
                            idx_subgrid = self.idxs[order_subgrid].index(idxs)
                            subgrid = self.grids[order_subgrid][idx_subgrid]
                            
                            # Find where to subtract the values
                            # Look for the values with the same idxs coordinates
                            level_grid_Xfilt = level_grid.X_DIM[:,list(idxs)]
                            subgrid_Xfilt = subgrid.X_DIM[:,list(idxs)]
                            for X,val in zip(subgrid_Xfilt,subgrid.ANOVA_HDMR_vals):
                                isEqual = Misc.almostEqualList(level_grid_Xfilt,X,self.tol)
                                level_grid.ANOVA_HDMR_vals[isEqual] -= val
    



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

import numpy as np
import scipy.special as scsp
import scipy.stats as stats
from matplotlib import pyplot as plt
from SpectralToolbox import Spectral1D
from UQToolbox import CutANOVA
import UQToolbox.RandomSampling as RS

DIM = 8
pp = Spectral1D.Poly1D(Spectral1D.JACOBI,[0.,0.])
polys = [pp for i in range(DIM)]
Ns = [6 for i in range(DIM)]
cut_order = 2
X_cut = np.zeros((1,len(polys)))

tol = 2. * np.spacing(1)

cut_HDMR = CutANOVA.CutHDMR(polys,Ns,cut_order,X_cut,tol)

def fun(X,params=None):
    a = np.asarray([0., 1.,4.5, 9., 99., 99., 99., 99.]);
    if len(X.shape) == 1:
        Y = np.prod( (np.abs(4.*X - 2.) + a)/(1. + a) );
    elif len(X.shape) == 2:
        Y = np.prod( (np.abs(4.*X - 2.) + a)/(1. + a) , 1);
    return Y

def transformFunc(X):
    # from [-1,1] to [0,1]
    return (X+1.)/2.

# Evaluate f on the cutHDMR grid
print "N. eval = %d" % np.sum( [scsp.binom(DIM,i) * Ns[0]**i for i in range(cut_order+1)] )
cut_HDMR.evaluateFun(fun,transformFunc)
print "End evaluation"

print "Start HDMR computation"
# Compute the cutHDMR
cut_HDMR.computeCutHDMR()

# Compute the ANOVA-HDMR
cut_HDMR.computeANOVA_HDMR()
print "End HDMR computation"

# Compute an estimate for the total variance (using Monte Carlo)
dists = [stats.uniform(0,1) for i in xrange(DIM)]
exp_lhc = RS.Experiments(fun,None,dists,False)
exp_lhc.sample(2000, method='lhc')
exp_lhc.run()
MC_vals = np.asarray(exp_lhc.get_samples())
MC_exp = np.asarray(exp_lhc.get_results())
MC_mean = np.mean(MC_exp)
MC_var = np.var(MC_exp)
plt.figure()
plt.subplot(2,1,1)
plt.plot(np.array([np.mean(MC_exp[:i]) for i in range(len(MC_exp))]))
plt.subplot(2,1,2)
plt.plot(np.array([np.var(MC_exp[:i]) for i in range(len(MC_exp))]))

# Compute individual variances
D = []
for level_grids in cut_HDMR.grids:
    D_level = []
    for grid in level_grids:
        D_level.append( np.dot(grid.ANOVA_HDMR_vals**2., grid.WF) )
    D.append(D_level)
Var_ANOVA = np.sum(D[1]) + np.sum(D[2])
print "TotVar/Var_Anova = %f" % (Var_ANOVA/MC_var)

# Compute Total variances per component
TV = np.zeros(DIM)
for idx in range(DIM):
    for level, level_idxs in enumerate(cut_HDMR.idxs):
        for j, idxs in enumerate(level_idxs):
            if idx in idxs:
                TV[idx] += D[level][j]
TS = TV/MC_var
print "N     TV       TS"
for i,grid in enumerate(cut_HDMR.grids[1]):
    print "%2d    %.4f   %.4f" % (i+1, TV[i], TS[i])

plt.figure()
plt.pie(TS/np.sum(TS),labels=["x%d"%(i+1) for i in range(DIM)])
plt.show(False)



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

__all__ = ['object_store','Driver']

import os
try:
    import cPickle as pkl
except ImportError:
    import pickle as pkl
import shutil
import sys
import warnings
import logging
import marshal, types
import time
import datetime
import numpy as np
from numpy import linalg as la
from numpy import random
from scipy import stats
from UQToolbox.sobol_lib import i4_sobol_generate
try:
    import mpi_map
    MPI_SUPPORT = True
except ImportError:
    MPI_SUPPORT = False

def object_store(path,obj):
    """ Used to store any object in the library.

    :param string path: path pointing to the location where to store the data
    :param object obj: a pickleable object

    """
    if os.path.isfile(path):
        # If the file already exists, make a safety copy
        shutil.copyfile(path, path+".old")
    ff = open(path,'wb')
    pkl.dump(obj,ff)
    ff.close()

class Driver(object):
    """ This class is devoted to the sampling from a multi dimensional distribution and the evaluation of a function `f` on it.
    
    :param function f: the function representing the experiment
    :param object params: parameters to be passed to the function ``f`` (pickable)
    :param list dists: list of distributions, instance of ``scipy.stats``
    :param bool marshal_f: whether to marshal the function ``f`` or not.
    :param str store_file: file path where to store the computed values
    """

    logger = logging.getLogger(__name__)
    logger.propagate = False
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("%(asctime)s %(levelname)s:%(name)s: %(message)s",
                                  "%Y-%m-%d %H:%M:%S")
    ch.setFormatter(formatter)
    logger.addHandler(ch)
        
    def __init__(self, f, params, marshal_f=True, store_file=""):

        self.f = None
        
        self.params = None
        self.f_code = None
        self.new_samples = []
        self.samples = []
        self.results = []
        self.store_file = ""
    
        self.serialize_list = ['new_samples','samples','results','f_code','params','store_file','serialize_list']
        
        self.set_f(f)
        self.set_params(params)
        self.store_file = store_file

    def __getstate__(self):
        return dict( [ (tag, getattr( self, tag )) for tag in self.serialize_list ] )

    def __setstate__(self,state,f = None, dists=None):
        for tag in state.keys():
            setattr(self, tag, state[tag])
        # Reset additional parameters
        if f == None: self.reset_f_marshal()
        else: self.set_f(f)
        
    def reset(self):
        self.samples = []
        self.new_samples = []
        self.results = []
    
    def set_f(self, f, marshal_f=True):
        self.f = f
        if self.f != None and marshal_f:
            self.f_code = marshal.dumps(self.f.__code__)
    
    def reset_f_marshal(self):
        if self.f_code != None:
            code = marshal.loads(self.f_code)
            self.f = types.FunctionType(code, globals(), "f")
        else:
            warnings.warn("UQToolbox.RandomSampling.Experiments: The Experiments has not function code to un-marshal. The function is undefined. Define it using UQToolbox.RandomSampling.Experiments.set_f", RuntimeWarning)
    
    def set_params(self, params):
        self.params = params
    
    def get_params(self):
        return self.params
        
    def get_new_samples(self):
        """
        :returns: the list of samples for which the experiment have not been evaluated yet.
        """
        return self.new_samples
    
    def get_samples(self):
        """
        :returns: the list of samples for which the experiment have been evaluated already.
        """
        return self.samples
    
    def get_results(self):
        """
        :returns: the list of results of the evaluation of ``f`` on the samples.
        """
        return self.results
    
    def store(self):
        if self.store_file != "":
            object_store(self.store_file, self)    
    
    def run (self, maxprocs=None, store_freq=None):
        """ Evaluate the function ``f`` on the samples.
        
        :param int maxprocs: the maximum number of processors available
        :param int store_freq: number of evaluations to be executed before storing. If ``maxprocs`` > 1 then the object is store every ``store_freq*maxprocs`` evaluations.
        """
        if maxprocs != None and not MPI_SUPPORT:
            warnings.warn("UQToolbox.RandomSampling.Experiments.run: MPI is not supported. The jobs will be run without it.", RuntimeWarning)
        
        if maxprocs == None or not MPI_SUPPORT:
            maxprocs = 1
        
        if store_freq == None:
            store_freq = len(self.new_samples)
        
        # Split the samples into chuncks of store_freq*maxprocs size
        chunk_size = store_freq*maxprocs
        ns = [ chunk_size ] * ( len(self.new_samples) // chunk_size )
        ns.append( len(self.new_samples) % chunk_size )
        for i in range(1,len(ns)): ns[i] += ns[i-1]
        ns.insert(0,0)
        slists = [ self.new_samples[ns[i]:ns[i+1]] for i in range(0, len(ns)-1 ) ]

        for slist in slists:
            start_time = time.time()
            if maxprocs == 1:
                for sample in slist:
                    self.results.append( self.f( sample, self.params ) )
            else:
                self.results.extend( mpi_map.mpi_map(self.f, slist, self.params, maxprocs) )
            stop_time = time.time()
            elapsed_time = stop_time-start_time
                        
            for sample in slist:
                self.samples.append(sample)
                self.new_samples.remove(sample)

            if len(slist) > 0:
                est_time = (elapsed_time/len(slist)) * len(self.new_samples)
                self.logger.info(
                    "Completed: %d/%d \t Est. Time: %s" % (
                        len(self.samples),
                        len(self.samples)+len(self.new_samples),
                        datetime.timedelta(seconds=est_time)
                    ))
                self.store()



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#


import warnings
import sys
import marshal, types
from functools import reduce

import numpy as np

from numpy import linalg as npla

from scipy import linalg as scla

from SpectralToolbox import Spectral1D
from SpectralToolbox import SpectralND

class GaussianField:
    """
    Provided a Covariance function and discretization points, it generates the Gaussian fields.

    :param function C: Covariance function C(x1,x2)
    """

    C = None
    L = None
    
    serialize_list = ['x','L']
    marshall_list = ['C']
    
    def __init__(self, C):
        self.C = C

    def __getstate__(self):
        dd = dict( [ (tag, getattr( self, tag )) for tag in self.serialize_list ] )
        dd['C_code'] = marshal.dumps(self.C.func_code)
        return dd
        
    def __setstate__(self, state):
        for tag in state.keys():
            if tag == 'C_code':
                code =  marshal.loads( state[tag] )
                self.C = types.FunctionType(code, globals(), "C")
            else:
                setattr(self, tag, state[tag])

    def fit(self, x):
        """
        Given some discretization points, it finds the Cholesky factorization to be used for the
        simulation of the random field
        
        :param np.ndarray x: array containing the discretization points
        """
        self.x = x
        X1,X2 = np.meshgrid(x,x)
        self.L = npla.cholesky( self.C(X1,X2) )
        self.nvars = len(self.x)
    
    def transform(self, xi):
        """
        Generates random fields from the input variables (Normally distributed)
        
        :param np.ndarray xi: 1 or 2 dimensional array containing the generating variables. If 2-dimensional, each row should contain a different sample.
        
        :returns: the realizations of the random field
        :rtype: np.ndarray
        """
        if xi.ndim == 1:
            xi = np.array([xi])
        return np.dot( self.L, xi.T ).T
        

class KLExpansion:
    """
    KLExpansion: Computes the 1D KL expansion of the covariance matrix C, using the Legendre-Gauss-Loabatto rule for the solution to the generalized eigenvalue problem A * u = lmb * M * u.
    The KLExpansion of d-dimensional random fields is obtained by tensor product.

    :param function C: Covariance function C(x1,x2)
    :param int d: dimension of the tensorized field
    :param list spans: list of tuples containing the spans spans of the coordinates of the domain of C(x1,x2)
    """

    C = None
    N = None
    d = None
    spans = None
    targetVar = None
    totVar = None
    
    x = None
    w = None
    eigvals = None
    eigvecs = None
    nvars = None
    tensor_valI = None
    order_valI = None

    # New points on which to interpolate the eigenvalues
    # These are auxiliary in order to speed up the computations
    xs_fit = None
    eigvecs_fit = None 
    multidim_eigvecsvals = None

    serialize_list = ['N','d','targetVar','totVar','x','w','eigvals','eigvecs','nvars','tensor_valI','order_valI','spans','xs_fit','eigvecs_fit','multidim_eigvecsvals']
    marshall_list = ['C']

    def __init__(self,C,d,spans):
        if isinstance(C,list):
            if len(C) == d:
                self.C = C
            else: raise TypeError("The length of C is not consistent with d")
        else:
            self.C = [C]*d
        self.d = d
        self.spans = spans

    def __getstate__(self):
        dd = dict( [ (tag, getattr( self, tag )) for tag in self.serialize_list ] )
        dd['C_code'] = [marshal.dumps(C.func_code) for C in self.C]
        return dd
        
    def __setstate__(self, state):
        for tag in state.keys():
            if tag == 'C_code':
                self.C = []
                if isinstance(state[tag],list):
                    # For new format with anisotropic covariances
                    for C_code in state[tag]:
                        code =  marshal.loads( C_code )
                        self.C.append(types.FunctionType(code, globals(), "C"))
                else:
                    code =  marshal.loads( state[tag] )
                    self.C.append(types.FunctionType(code, globals(), "C"))
            else:
                setattr(self, tag, state[tag])
    
    def scale(self,i,x):
        """ Scales values x in [-1,1] to ``self.spans[i]``.
        """
        return ((x+1.)/2. + self.spans[i][0]) * (self.spans[i][1] - self.spans[i][0])
    
    def invscale(self,i,x):
        """ Scales values x in ``self.spans[i]`` to [-1,1].
        """
        return (x-self.spans[i][0])/(self.spans[i][1]-self.spans[i][0]) * 2. - 1.

    def fit(self,N,targetVar=.95):
        """
        Sets up the KL-expansion eigenvalues and eigenvectors for the required tolerance.

        :param list N: list of number of discretization points per dimension
        :param float targetVar: total variance to be represented by the output KL-expansion
        """
        self.N = N
        self.targetVar = targetVar
        
        self.x = []
        self.w = []

        self.eigvals = []
        self.eigvecs = []
        
        for d in range(self.d):
            # Lagrange polynomials....
            poly = Spectral1D.generate(Spectral1D.JACOBI,[0.,0.])
            (xI,wI) = poly.GaussLobattoQuadrature(2 * self.N[d])
            wI /= np.sum(wI)
            self.x.append(xI)
            self.w.append(wI)

            xSc = self.scale(d,xI)
            X1,X2 = np.meshgrid(xSc,xSc)

            M = np.diag(wI)
            K = np.dot(M, np.dot(self.C[d](X1,X2),M ) )

            ''' Solve generalized eigenvalue problem '''
            (valI,vecI) = scla.eigh(K,M)
            # Normalize vectors and transfer normalization constant to eigenvalues
            nrm = np.diag( np.dot( np.multiply( wI, vecI.T ) , vecI ))
            vecI = np.multiply( 1./nrm , vecI )
            valI *= nrm
            # Sort in ascending order
            order = valI.argsort()[::-1]
            valI = valI[order]
            vecI = vecI[:,order]
            # Truncate to N
            valI = np.real(valI[:self.N[d]])
            vecI = np.real(vecI[:,:self.N[d]])

            TOTVAR = np.sum( valI )
            if TOTVAR > 1. + 1e-10:
                warnings.warn("UQToolbox.ModelReduction.KLExpansion: The total variance expressed by the KL-expansion exceed 1. TOTVAR = %f" % TOTVAR ,RuntimeWarning)
            
            self.eigvals.append(valI)
            self.eigvecs.append(vecI)

        tensor_valI = reduce(np.multiply, np.ix_( *self.eigvals ))
        order_valI = np.vstack(np.unravel_index(np.argsort(tensor_valI,axis=None)[::-1],(tensor_valI.shape) ))
        self.tensor_valI = tensor_valI
        self.order_valI = order_valI

        ''' Compute variance represented by the successive modes '''
        j = 0
        self.totVar = 0.
        while (self.totVar < self.targetVar) and (j < order_valI.shape[1]):
            tmp = 1.
            for i in range(self.d):
                tmp *= self.eigvals[i][order_valI[i,j]] * \
                    np.dot( wI * self.eigvecs[i][:,order_valI[i,j]], self.eigvecs[i][:,order_valI[i,j]])
            self.totVar += tmp
            j += 1

        if self.totVar < self.targetVar:
            warnings.warn("UQToolbox.ModelReduction.KLExpansion: length of the expansion not sufficient to resolve the target variance:\n KL-expansion: Target Variance      = %f \n KL-expansion: Represented Variance = %f" % (self.targetVar, self.totVar), RuntimeWarning)
        
        self.nvars = j        
    
    def fit_grid(self, xs):
        """ Used to pre-fit the eigenfunctions to a set of grid points.
        :params list xs: list of d coordinates for each direction
        """

        shape = tuple([len(x) for x in xs])
        # Scale the values to the interval [-1,1]
        self.xs_fit = [ self.invscale(i,x) for i,x in enumerate(xs) ]

        # Extrapolate values of eigenvalues at points x
        self.eigvecs_fit = []
        for j in range(self.d):
            if len(self.x[j]) == len(self.xs_fit[j]) \
               and npla.norm( self.x[j] - self.xs_fit[j] ) <= np.spacing(1):
                self.eigvecs_fit.append( self.eigvecs[j] )
            else:
                self.eigvecs_fit.append( Spectral1D.LagrangeInterpolate(self.x[j], self.eigvecs[j], self.xs_fit[j]) )
    
        self.multidim_eigvecsvals = []
        for j in range(self.nvars):
            tmp = np.ones(1)
            for i in range(self.d):
                tmp = np.kron(tmp, np.sqrt(self.eigvals[i][self.order_valI[i,j]]) * self.eigvecs_fit[i][:,self.order_valI[i,j]] )
            self.multidim_eigvecsvals.append(tmp)
        self.multidim_eigvecsvals = np.array(self.multidim_eigvecsvals).T

    def transform(self,xs,xi):
        """
        :params list xs: list of d coordinates for each direction
        :params list xi: list of 1 dimensional arrays of ``self.nvars`` values xi or array of ``self.nvars`` values.
        """
        
        shape = tuple([len(x) for x in xs])
        # Scale the values to the interval [-1,1]
        xs = [ self.invscale(i,x) for i,x in enumerate(xs) ]

        xi_islist = True
        if isinstance(xi,np.ndarray):
            xi = [xi]
            xi_islist = False
        NR = len(xi)

        if np.any( [len(x) for x in xi] != [self.nvars]*NR ):
            raise RuntimeError("UQToolbox.ModelReduction.KLExpansion: the number of coefficients is not consistent with the length of the expansion.")

        # Extrapolate values of eigenvalues at points x
        eigvecs = []
        PREFIT = False
        for j in range(self.d):
            if len(self.x[j]) == len(xs[j]) and npla.norm( self.x[j] - xs[j] ) <= np.spacing(1):
                eigvecs.append( self.eigvecs[j] )
            elif self.xs_fit != None and len(self.xs_fit[j]) == len(xs[j]) and npla.norm( self.xs_fit[j] - xs[j] ) <= np.spacing(1):
                PREFIT = True
            else:
                eigvecs.append( Spectral1D.LagrangeInterpolate(self.x[j], self.eigvecs[j], xs[j]) )
        
        # Compute the KL-expansion for xi at the self.x points (Pascal's triangle)
        if PREFIT:
            u = [ np.dot( self.multidim_eigvecsvals, xi[i] ) for i in range(NR) ]
        else:
            u = [ 0. for i in range(NR) ]
            for j in range(self.nvars):
                tmp = np.ones(1)
                for i in range(self.d):
                    tmp = np.kron(tmp, np.sqrt(self.eigvals[i][self.order_valI[i,j]]) * eigvecs[i][:,self.order_valI[i,j]] )
                for i in range(NR):
                    u[i] += tmp * xi[i][j]
        
        for i in range(NR): u[i] = np.reshape(u[i],shape)

        if xi_islist:
            return u
        else:
            return u[0]

def KLExpansion1D(C,N,span,targetVar=0.95):
    """
    :note: Deprecated. Use the class ``KLExpansion`` instead

    KLExpansion: Computes the 1D KL expansion of the covariance matrix C, using the
       Legendre-Gauss-Loabatto rule for the solution to the generalized eigenvalue 
       problem A * u = lmb * M * u

    :param function C: Covariance function C(x1,x2)
    :param int N: number of discretization points    
    :param tuple span: span of the domain of C(x1,x2)
    :param float targetVar: total variance to be represented by the output KL-expansion    
    """
    warnings.warn("UQToolbox.ModelReduction.KLExpansion1D: This function is deprecated. Use the UQToolbox.ModelReduction.KLExpansion class instead.", DeprecationWarning)
    
    KLE = KLExpansion(C,1,[span])
    KLE.fit([N],targetVar)
    return ([KLE.scale(i,x) for i,x in enumerate(KLE.x)], KLE.eigvals, KLE.eigvecs, KLE.totVar, KLE.nvars)



# -*- coding: utf-8 -*-

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#

import sys
import os
import warnings
import logging
import marshal, types
import time
import datetime
import numpy as np
from numpy import linalg as la
from numpy import random
from scipy import stats
from UQToolbox.sobol_lib import i4_sobol_generate
from UQToolbox import object_store, Driver
try:
    import mpi_map
    MPI_SUPPORT = True
except ImportError:
    MPI_SUPPORT = False

class MultiDimDistribution():
    """ Tensor construction of a multidimensional distribution
    
    :param list dists: list of distributions along each dimension.
    """

    dists = []
    
    def __init__(self,dists):
        self.dists = dists;
    
    def rvs(self,size):
        """ Samples ``size`` realizations from the multidimensional distribution. Equivalent to Monte Carlo sampling.
        
        :param int size: number of samples
        :returns: list of samples
        """
        samples = np.zeros( (size, len(self.dists)) )
        for i in range(len(self.dists)):
            samples[:,i] = self.dists[i].rvs(size)
        return list(samples)
    
    def lhc(self,size):
        """ Samples ``size`` realizations from the multidimensional distribution using the Latin Hyper Cube method.
        
        :param int size: number of samples
        :returns: list of samples
        """
        d = len(self.dists)
        
        XX = np.zeros((size,d))
        for i in range(0,size):
            XX[i,:] = stats.uniform(loc=(i*1./size),scale=1./size).rvs(size=d)

        P = np.zeros((size,d),dtype=np.int)
        for i in range(0,d):
            P[:,i] = np.arange(0,size)
            random.shuffle(P[:,i])

        for i in range(0,d):
            XX[:,i] = XX[P[:,i],i]

        ''' Convert from uniform to self.dists '''
        udist = stats.uniform(0.,1.);
        if self.dists != None and len(self.dists) == d:
            for i in range(d):
                XX[:,i] = self.dists[i].ppf(udist.cdf(XX[:,i]))

        return list(XX)
    
    def sobol(self,size,skip=None):
        """ Samples ``size`` realizations from the multidimensional distribution using the Sobol sequence for Quasi Monte Carlo method.
        
        :param int size: number of samples
        :returns: list of samples
        """
        if skip == None:
            dim = len(self.dists)
            skip = int( np.random.uniform(2**np.ceil(np.log2(dim+1)), 2**(np.ceil(np.log2(dim+1))+1)) )
        unifSamples = i4_sobol_generate(len(self.dists),size,skip); 
        samples = np.zeros(unifSamples.T.shape);
        for i in range(0,len(self.dists)):
            samples[:,i] = self.dists[i].ppf(unifSamples[i,:])
        
        return list(samples)
    
class RandomExperiments(Driver):
    """ This class is devoted to the sampling from a multi dimensional distribution and the evaluation of a function `f` on it.
    
    :param function f: the function representing the experiment
    :param object params: parameters to be passed to the function ``f`` (pickable)
    :param list dists: list of distributions, instance of ``scipy.stats``
    :param bool marshal_f: whether to marshal the function ``f`` or not.
    :param str store_file: file path where to store the computed values
    """

    logger = logging.getLogger(__name__)
    logger.propagate = False
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("%(asctime)s %(levelname)s:%(name)s: %(message)s",
                                  "%Y-%m-%d %H:%M:%S")
    ch.setFormatter(formatter)
    logger.addHandler(ch)
        
    def __init__(self, f, params, dists, marshal_f=True, store_file=""):

        super(RandomExperiments,self).__init__(f, params, marshal_f, store_file)
        
        self.dists = None
        self.mdist = None
        
        self.set_dists(dists)

    def __getstate__(self):
        return super(RandomExperiments,self).__getstate__()

    def __setstate__(self,state,f = None, dists=None):
        super(RandomExperiments,self).__setstate__( state, f )
        if dists != None:
            self.set_dists( dists )
            
    def set_dists(self, dists):
        self.dists = dists
        self.mdist = MultiDimDistribution(dists)
    
    def get_mdist(self):
        return self.mdist
    
    def sample(self, size, method='mc', append=True):
        """ Sample from the multidimensional distribution defined by the ``dists``.
        
        :param int size: the size of the sample set
        :param str method: the sampling method to be used. Options are: 'mc' for Monte Carlo, 'lhc' for Latin Hyper Cube, 'sobol' for Quasi Monte Carlo with Sobol sequence.
        :param bool append: if True, the new samples are appended to the old samples. Using run() will apply the experiments on only the new samples. If False, the old samples and old results are discarded. This is possible only for methods 'mc' and 'lhc'.

        :note: If the Experiments object has been reloaded using pickle, the ``dists`` parameters should be reset using ``set_dists``.
        """
        if self.mdist == None:
            raise RuntimeError("UQToolbox.RandomSampling.Experiments.sample: No distribution set. Use UQToolbox.RandomSampling.Experiments.set_dists to set the distribution.")
        
        if append == False or method == 'sobol':
            self.reset()
        
        if method == 'mc':
            self.new_samples.extend( self.mdist.rvs(size) )
        elif method == 'lhc':
            self.new_samples.extend( self.mdist.lhc(size) )
        elif method == 'sobol':
            self.new_samples.extend( self.mdist.sobol(size) )
        else:
            raise RuntimeError("UQToolbox.RandomSampling.Experiments.sample: The proposed method is not available.")

Experiments = RandomExperiments

def run_experiments(f,samples,params,paramUpdate,action):
    """
    Compute the Experiments f on the samples.
    The implementation uses MPI for parallel computations.

    :param function f: experiment function handle. Signature: f( params )
    :param samples: nd-array with the set of samples grouped by the first dimension
    :param params: set of parameters to be passed to the experiment
    :param function action: post processing action

    :returns: Array of computed values, ordered by the first dimension of the array.

    ..deprecated:: 0.1.5
    """
    warnings.warn("UQToolbox.RandomSampling.run_experiments: Deprecated function", DeprecationWarning)
    
    def iterF(f,samples,params,paramUpdate,action):
        sols = []
        for i in range(0,samples.shape[0]):
            params = paramUpdate(params,samples[i])
            sol = f(params)
            print("Proc %d run %d/%d" % (myrank,i+1,len(samples)))
            sols = action(sols,sol)
        return sols
    
    comm = MPI.COMM_WORLD
    nprocs = comm.Get_size()
    myrank = comm.Get_rank()
    
    if myrank == 0:
        # Split the input array
        splittedSamples = np.array_split(samples,nprocs)
        startTime = time.clock()
    else:
        splittedSamples = None
    
    samplesPart = comm.scatter(splittedSamples,root=0)
    
    splittedSolutions = iterF(f,samplesPart,params,paramUpdate,action)
    
    solutionsList = comm.gather(splittedSolutions)
    
    if myrank == 0:
        # Reassemble post processing data
        # To be fixed for MPI!!!!!! Use the proper action..
        solutions = [inner for outer in solutionsList for inner in outer]

        stopTime = time.clock()
        print("Elapsed Time: %f s" % (stopTime-startTime))
        
        return solutions

def MonteCarlo(dists,N,experiment,params,paramUpdate,postProc):
    """
    Run Monte Carlo Simulations
    
    ..deprecated:: 0.1.5
    """
    warnings.warn("UQToolbox.RandomSampling.MonteCarlo: Deprecated function", DeprecationWarning)

    mdd = MultiDimDistribuion(dists);
    samples = mdd.rvs(N)
    return (samples,run_experiments(experiment, samples, params, paramUpdate, postProc))

def QuasiMonteCarlo(dists,N,experiment,params,paramUpdate,postProc,skip=None):
    """
    Run Quasi Monte Carlo Simulations

    ..deprecated:: 0.1.5
    """
    warnings.warn("UQToolbox.RandomSampling.QuasiMonteCarlo: Deprecated function", DeprecationWarning)

    # Generate uniformly distributed samples using Sobol sequence
    if skip == None:
        dim = len(dists)
        skip = int( np.random.uniform(2**np.ceil(np.log2(dim+1)), 2**(np.ceil(np.log2(dim+1))+1)) )
    unifSamples = i4_sobol_generate(len(dists),N,skip); 
    samples = np.zeros(unifSamples.T.shape);
    for i in range(0,len(dists)):
        samples[:,i] = dists[i].ppf(unifSamples[i,:])
    return (samples,run_experiments(experiment, samples, params, paramUpdate, postProc), skip)

def lhc(N,d,dists=None):
    """
    Latin Hyper Cube

    ..deprecated:: 0.1.5
    """
    warnings.warn("UQToolbox.RandomSampling.lhc: Deprecated function", DeprecationWarning)

    XX = np.zeros((N,d))
    for i in range(0,N):
        XX[i,:] = stats.uniform(loc=(i*1./N),scale=1./N).rvs(size=d)
    
    P = np.zeros((N,d),dtype=np.int)
    for i in range(0,d):
        P[:,i] = np.arange(0,N)
        random.shuffle(P[:,i])
    
    for i in range(0,d):
        XX[:,i] = XX[P[:,i],i]
    
    ''' Convert from uniform to dists '''
    udist = stats.uniform(0.,1.);
    if dists != None and len(dists) == d:
        for i in range(d):
            XX[:,i] = dists[i].ppf(udist.cdf(XX[:,i]))
    
    return XX

def LatinHyperCube(dists,N,experiment,params,paramUpdate,postProc):
    """
    Run Latin Hyper Cube Simulations
    
    ..deprecated:: 0.1.5
    """
    warnings.warn("UQToolbox.RandomSampling.LatinHyperCube: Deprecated function", DeprecationWarning)

    samples = lhc(N,len(dists),dists)
    return (samples,run_experiments(experiment, samples, params, paramUpdate, postProc))



#!/usr/bin/env python

#
# This file is part of UQToolbox.
#
# UQToolbox is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UQToolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with UQToolbox.  If not, see <http://www.gnu.org/licenses/>.
#
# DTU UQ Library
# Copyright (C) 2014 The Technical University of Denmark
# Scientific Computing Section
# Department of Applied Mathematics and Computer Science
#
# Author: Daniele Bigoni
#


import os.path
import sys, re
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
import pip

global include_dirs
include_dirs = []

deps_list = ['numpy', 'orthpol', 'scipy',
             'Sphinx', 'sphinxcontrib-bibtex',
             'SpectralToolbox', 'prettytable']
if "--without-mpi4py" in sys.argv:
    idx = sys.argv.index('--without-mpi4py')
    del sys.argv[idx]
else:
    deps_list.extend(['mpi4py', 'mpi_map'])

def deps_install():
    for package in deps_list:
        print("[DEPENDENCY] Installing %s" % package)
        try:
            pip.main(['install', '--no-binary', ':all:', '--upgrade', package])
        except Exception as e:
            print("[Error] Unable to install %s using pip." +
                  "Please read the instructions for " +
                  "manual installation... Exiting" % package)
            raise e
            exit(2)

class UQToolbox_install(install):
    def run(self):
        deps_install()
        import numpy as np
        include_dirs.append(np.get_include())
        install.run(self)

class UQToolbox_develop(develop):
    def run(self):
        deps_install()
        import numpy as np
        include_dirs.append(np.get_include())
        develop.run(self)

local_path = os.path.split(os.path.realpath(__file__))[0]
version_file = os.path.join(local_path, 'UQToolbox/_version.py')
version_strline = open(version_file).read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, version_strline, re.M)
if mo:
    version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (version_file,))

setup(name = "UQToolbox",
      version = version,
      packages=find_packages(),
      include_package_data=True,
      scripts=[],
      url="http://www2.compute.dtu.dk/~dabi/",
      author = "Daniele Bigoni",
      author_email = "dabi@dtu.dk",
      license="COPYING.LESSER",
      description="Tools for Uncertainty Quantification",
      long_description=open("README.txt").read(),
      zip_safe = False,                      # For Debug purposes
      cmdclass={'install': UQToolbox_install,
                'develop': UQToolbox_develop},
      include_dirs=include_dirs
      )
